/**************************************************************************

   Fotoxx      edit photos and manage collections

   Copyright 2007 2008 2009 2010 2011 2012 2013 Michael Cornelison
   Source URL: http://kornelix.com/fotoxx
   Contact: kornelix2@gmail.com
   
   This program is free software: you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation, either version 3 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program. If not, see http://www.gnu.org/licenses/.

***************************************************************************/

#define EX extern                                                          //  enable extern declarations
#include "fotoxx.h"


/**************************************************************************
   Fotoxx image edit - enhance functions
***************************************************************************/


/**************************************************************************/

//  automatic image tuneup without user guidance

float       voodoo_brdist[256];                                            //  pixel count per brightness level
int         voodoo_01, voodoo_99;                                          //  1% and 99% brightness levels (0 - 255)
void        voodoo_distribution();                                         //  compute brightness distribution
void *      voodoo_thread(void *);
editfunc    EFvoodoo;


void m_voodoo(GtkWidget *, cchar *)                                        //  new v.12.06
{
   F1_help_topic = "voodoo";

   EFvoodoo.funcname = "voodoo";
   EFvoodoo.Fprev = 0;                                                     //  no preview
   EFvoodoo.Farea = 1;                                                     //  select area ignored
   EFvoodoo.threadfunc = voodoo_thread;
   if (! edit_setup(EFvoodoo)) return;                                     //  setup edit

   voodoo_distribution();                                                  //  compute brightness distribution
   signal_thread();                                                        //  start update thread
   edit_done();                                                            //  edit done
   return;
}


//  compute brightness distribution for image

void voodoo_distribution()
{
   int         px, py, ii;
   int         ww = E1pxm->ww, hh = E1pxm->hh;
   float       bright1;
   float       *pix1;

   for (ii = 0; ii < 256; ii++)                                            //  clear brightness distribution data
      voodoo_brdist[ii] = 0;

   for (py = 0; py < hh; py++)                                             //  compute brightness distribution
   for (px = 0; px < ww; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);
      bright1 = pixbright(pix1);
      voodoo_brdist[int(bright1)]++;
   }
   
   for (ii = 1; ii < 256; ii++)                                            //  cumulative brightness distribution
      voodoo_brdist[ii] += voodoo_brdist[ii-1];                            //   0 ... (ww * hh)
   
   voodoo_01 = 0;
   voodoo_99 = 255;

   for (ii = 0; ii < 256; ii++)                                            //  compute index values (0 - 255)
   {                                                                       //    for darkest 1% and brightest 1%
      if (voodoo_brdist[ii] < 0.01 * ww * hh) voodoo_01 = ii;
      if (voodoo_brdist[ii] < 0.99 * ww * hh) voodoo_99 = ii;
   }
   
   for (ii = 0; ii < 256; ii++)
      voodoo_brdist[ii] = voodoo_brdist[ii]                                //  multiplier per brightness level 
                    / (ww * hh) * 256.0 / (ii + 1);                        //  ( = 1 for a flat distribution)
   return;
}


//  thread function - use multiple working threads

void * voodoo_thread(void *)
{
   void  * voodoo_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(voodoo_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++; 
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}

void * voodoo_wthread(void *arg)                                           //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py;
   float       *pix1, *pix3;
   float       bright1, bright2, bright3, cmax;
   float       red1, green1, blue1, red3, green3, blue3;
   float       flat1 = 0.3;                                                //  strength of distribution flatten
   float       flat2 = 1.0 - flat1;
   float       sat1 = 0.4, sat2;                                           //  strength of saturation increase
   float       f1, f2;
   float       expand = 256.0 / (voodoo_99 - voodoo_01 + 1);               //  brightness range expander
   
   for (py = index; py < E1pxm->hh; py += Nwt)                             //  voodoo brightness distribution
   for (px = 0; px < E1pxm->ww; px++)
   {
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];

      bright2 = 0.25 * red1 + 0.65 * green1 + 0.10 * blue1;                //  input brightness, 0 - 256
      bright2 = (bright2 - voodoo_01) * expand;                            //  expand to clip low / high 1% 
      if (bright2 < 0) bright2 = 0;
      if (bright2 > 255) bright2 = 255;

      bright1 = voodoo_brdist[int(bright2)];                               //  factor for flat output brightness
      bright3 = flat1 * bright1 + flat2;                                   //  attenuate

      f1 = (256.0 - bright2) / 256.0;                                      //  bright2 = 0 - 256  >>  f1 = 1 - 0
      f2 = 1.0 - f1;                                                       //  prevent banding in bright areas
      bright3 = f1 * bright3 + f2;                                         //  tends to 1.0 for brighter pixels

      red3 = red1 * bright3;                                               //  blend new and old brightness
      green3 = green1 * bright3;
      blue3 = blue1 * bright3;
      
      bright3 = 0.333 * (red3 + green3 + blue3);                           //  mean color brightness
      sat2 = sat1 * (256.0 - bright3) / 256.0;                             //  bright3 = 0 - 256  >>  sat2 = sat1 - 0
      red3 = red3 + sat2 * (red3 - bright3);                               //  amplified color, max for dark pixels
      green3 = green3 + sat2 * (green3 - bright3);
      blue3 = blue3 + sat2 * (blue3 - bright3);
      
      if (red3 < 0) red3 = 0;                                              //  stop possible underflow
      if (green3 < 0) green3 = 0;
      if (blue3 < 0) blue3 = 0;

      cmax = red3;                                                         //  stop overflow, keep color balance
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      if (cmax > 255.9) {
         cmax = 255.9 / cmax;
         red3 = red3 * cmax;
         green3 = green3 * cmax;
         blue3 = blue3 * cmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  brightness / color / contrast adjustment

void  tune_curve_update(int spc);                                          //  curve update callback function
void  tune_update_image();                                                 //  process all pixels
void  tune_update_pixel(int px, int py);                                   //  process one pixel
void  *tune_thread(void *);                                                //  thread process for all pixels

int      tune_spc;                                                         //  current spline curve 1-5
int      tune_spc_moved[6];                                                //  tracks which curves changed

float    tune_pow[10000];                                                  //  pow(2,x) for x = -5.0 to +5.0
int      tune_ftf = 1;
int      tune_updated;

editfunc    EFtune;


void m_tune(GtkWidget *, cchar *)                                          //  menu function
{
   int   tune_dialog_event(zdialog *zd, cchar *event);

   cchar    *title = ZTX("Adjust Brightness and Color");
   F1_help_topic = "tune";
   
   EFtune.funcname = "bright-color";
   EFtune.Fprev = 1;                                                       //  use preview
   EFtune.Farea = 2;                                                       //  select area usable
   EFtune.menufunc = m_tune;                                               //  restart allowed                 v.13.04
   EFtune.threadfunc = tune_thread;
   if (! edit_setup(EFtune)) return;                                       //  setup edit
   
   if (tune_ftf) {
      tune_ftf = 0;                                                        //  pre-compute pow(2,x)
      for (int ii = 0; ii < 10000; ii++)                                   //   for x = -5.0 to +4.999
         tune_pow[ii] = pow(2,(ii/1000.0 - 5.0));
   }
   
/***
            Adjust Brightness and Color
       ____________________________________________
      |                                            |
      |                                            |
      |                                            |
      |           curve drawing area               |
      |                                            |
      |                                            |
      |____________________________________________|
       darker areas                   lighter areas

      [+++] [---] [+ -] [- +] [+-+] [-+-]
      [x] small-steps  Curve File: [ Open ] [ Save ]
      (o) brightness       [reset 1]  [reset all]
      (o) saturation       [histogram]
      (o) red
      (o) green
      (o) blue
                                   [done] [cancel]
***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFtune.zd = zd;

   zdialog_add_widget(zd,"frame","fr1","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hba","dialog");
   zdialog_add_widget(zd,"label","labda","hba",Bdarker,"space=5");
   zdialog_add_widget(zd,"label","space","hba",0,"expand");
   zdialog_add_widget(zd,"label","labba","hba",Blighter,"space=5");

   zdialog_add_widget(zd,"hbox","hbb","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","b +++","hbb"," +++ ");
   zdialog_add_widget(zd,"button","b ---","hbb","‒ ‒ ‒");
   zdialog_add_widget(zd,"button","b +-", "hbb"," + ‒ ");
   zdialog_add_widget(zd,"button","b -+", "hbb"," ‒ + ");
   zdialog_add_widget(zd,"button","b +-+","hbb","+ ‒ +");
   zdialog_add_widget(zd,"button","b -+-","hbb","‒ + ‒");

   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=5");
   zdialog_add_widget(zd,"check","smallstep","hbcf",ZTX("small-steps"));
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen);
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"vbox","vb21","hb2");
   zdialog_add_widget(zd,"vbox","vb22","hb2");
   zdialog_add_widget(zd,"radio","radbri","vb21",Bbrightness);
   zdialog_add_widget(zd,"radio","radsat","vb21",ZTX("Saturation"));
   zdialog_add_widget(zd,"radio","radR","vb21",Bred);
   zdialog_add_widget(zd,"radio","radG","vb21",Bgreen);
   zdialog_add_widget(zd,"radio","radB","vb21",Bblue);

   zdialog_add_widget(zd,"hbox","hbrs","vb22",0,"space=5");
   zdialog_add_widget(zd,"label","space","hbrs",0,"space=10");
   zdialog_add_widget(zd,"button","reset1","hbrs",ZTX(" reset 1 "));
   zdialog_add_widget(zd,"button","resetA","hbrs",ZTX("reset all"));
   zdialog_add_widget(zd,"hbox","hbhist","vb22",0,"space=6");
   zdialog_add_widget(zd,"label","space","hbhist",0,"space=10");
   zdialog_add_widget(zd,"button","histo","hbhist",Bhistogram);
   
   GtkWidget *frame = zdialog_widget(zd,"fr1");                            //  setup for curve editing
   spldat *sd = splcurve_init(frame,tune_curve_update);
   EFtune.curves = sd;

   sd->Nscale = 3;                                                         //  y-scale lines at 1.0 EV intervals
   sd->xscale[0][0] = 0.00;                                                //  x and y values
   sd->yscale[0][0] = 0.25;
   sd->xscale[1][0] = 1.00;
   sd->yscale[1][0] = 0.25;
   sd->xscale[0][1] = 0.00;
   sd->yscale[0][1] = 0.50;
   sd->xscale[1][1] = 1.00;
   sd->yscale[1][1] = 0.50;
   sd->xscale[0][2] = 0.00;
   sd->yscale[0][2] = 0.75;
   sd->xscale[1][2] = 1.00;
   sd->yscale[1][2] = 0.75;

   for (int spc = 0; spc < 6; spc++)                                       //  setup 6 spline curves
   {                                                                       //  no. 0 is active curve
      sd->vert[spc] = 0;                                                   //  all curves are horizontal
      sd->nap[spc] = 3;                                                    //  curves 1-5 are copied to 0 when active
      sd->apx[spc][0] = 0.01;
      sd->apy[spc][0] = 0.5;
      sd->apx[spc][1] = 0.5;
      sd->apy[spc][1] = 0.5;
      sd->apx[spc][2] = 0.99;
      sd->apy[spc][2] = 0.5;
      splcurve_generate(sd,spc);
      tune_spc_moved[spc] = 0;
   }

   sd->Nspc = 1;                                                           //  only one at a time is active
   tune_spc = 1;                                                           //  default curve = brightness
   tune_updated = 0;

   zdialog_stuff(zd,"radbri",1);                                           //  stuff default selection
   
   zdialog_resize(zd,0,500);
   zdialog_run(zd,tune_dialog_event,"save");                               //  run dialog - parallel
   return;
}


//  dialog event and completion callback function

int tune_dialog_event(zdialog *zd, cchar *event)
{
   int         Fupdate = 0;
   int         ii, jj, nn;
   static int  smallstep = 0;
   float       step, step2, step4;
   float       px, py;
   spldat      *sd = EFtune.curves;
   spldat      sdtemp;
   
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 1;
   }
   
   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }

   if (strEqu(event,"reset"))                                              //  reset dialog controls
      event = "resetA";

   if (strEqu(event,"restart"))                                            //  new image                    v.13.04
      if (tune_updated) Fupdate++;

   if (strEqu(event,"blendwidth")) tune_update_image();
   if (strEqu(event,"histo")) m_histogram(0,0);                            //  popup brightness histogram

   if (strnEqu(event,"rad",3)) {                                           //  new choice of curve
      ii = strcmpv(event,"radbri","radsat","radR","radG","radB",null);
      tune_spc = ii;
      sd->nap[0] = sd->nap[ii];                                            //  copy active curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      Fupdate++;
   }

   if (strEqu(event,"smallstep"))                                          //  toggle [+++] etc. step size
      zdialog_fetch(zd,"smallstep",smallstep);

   ii = tune_spc;                                                          //  current active curve

   if (strnEqu(event,"b ",2))                                              //  button [+++] etc. - move entire curve
   {
      step = 0.025;                                                        //  1/40
      if (smallstep) step = step * 0.3333;
      step2 = 2 * step;
      step4 = 4 * step;

      for (jj = 0; jj < sd->nap[0]; jj++) {                                //  12.04
         px = sd->apx[0][jj];
         py = sd->apy[0][jj];
         if (strEqu(event,"b +++")) py += step;                            //  range is 40 steps
         if (strEqu(event,"b ---")) py -= step;                            //   = -2.0 to +2.0 EV
         if (strEqu(event,"b +-"))  py += step - step2 * px;               //  normal steps are 0.1 EV or 0.03 OD
         if (strEqu(event,"b -+"))  py -= step - step2 * px;               //  small steps are 1/3 as big
         if (strEqu(event,"b +-+")) py -= step - step4 * fabsf(px-0.5);
         if (strEqu(event,"b -+-")) py += step - step4 * fabsf(px-0.5);
         if (py > 1) py = 1;
         if (py < 0) py = 0;
         sd->apy[0][jj] = py;
      }
      tune_spc_moved[ii] = 1;
      Fupdate++;
   }

   if (strEqu(event,"reset1")) {                                           //  reset current curve
      sd->nap[0] = 3;
      sd->apx[0][0] = 0.01;                                                //  3 anchor points, flatline
      sd->apy[0][0] = 0.5;
      sd->apx[0][1] = 0.5;
      sd->apy[0][1] = 0.5;
      sd->apx[0][2] = 0.99;
      sd->apy[0][2] = 0.5;
      Fupdate++;
      tune_spc_moved[ii] = 0;
   }
   
   if (strEqu(event,"resetA")) 
   {
      for (jj = 0; jj < 6; jj++) {                                         //  reset all curves
         sd->nap[jj] = 3;
         sd->apx[jj][0] = 0.01;
         sd->apy[jj][0] = 0.5;
         sd->apx[jj][1] = 0.5;
         sd->apy[jj][1] = 0.5;
         sd->apx[jj][2] = 0.99;
         sd->apy[jj][2] = 0.5;
         splcurve_generate(sd,jj);                                         //  regenerate all
         tune_spc_moved[jj] = 0;
      }
      splcurve_generate(sd,0);                                             //  regenerate curve 0           v.13.04
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      edit_reset();                                                        //  reset image mods
      tune_updated = 0;
   }

   if (strEqu(event,"load"))                                               //  load 5 saved curves
   {
      sdtemp.Nspc = 5;
      sdtemp.drawarea = 0;
      nn = splcurve_load(&sdtemp);
      if (nn != 1) return 0;

      for (ii = 0; ii < 5; ii++) {
         for (jj = 0; jj < sdtemp.nap[ii]; jj++) {
            sd->apx[ii+1][jj] = sdtemp.apx[ii][jj];
            sd->apy[ii+1][jj] = sdtemp.apy[ii][jj];
         }
         splcurve_generate(sd,ii+1);
         tune_spc_moved[ii+1] = 1;
      }

      ii = tune_spc;
      sd->nap[0] = sd->nap[ii];                                            //  copy active curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      Fupdate++;
   }

   if (strEqu(event,"save"))                                               //  save 5 curves to file
   {
      sdtemp.Nspc = 5;
      for (ii = 0; ii < 5; ii++) {
         sdtemp.nap[ii] = sd->nap[ii+1];
         for (jj = 0; jj < sd->nap[ii+1]; jj++) {
            sdtemp.apx[ii][jj] = sd->apx[ii+1][jj];
            sdtemp.apy[ii][jj] = sd->apy[ii+1][jj];
         }
      }
      splcurve_save(&sdtemp);
   }

   if (Fupdate)                                                            //  curve has changed
   {
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      
      ii = tune_spc;                                                       //  active curve
      sd->nap[ii] = sd->nap[0];                                            //  copy curve 0 to active curve
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[ii][jj] = sd->apx[0][jj];
         sd->apy[ii][jj] = sd->apy[0][jj];
      }
      for (jj = 0; jj < 1000; jj++)
         sd->yval[ii][jj] = sd->yval[0][jj];

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      tune_update_image();                                                 //  trigger image update
   }
   
   return 1;
}


//  this function is called when curve 0 is edited using mouse

void tune_curve_update(int)
{
   int      ii = tune_spc, jj;
   spldat   *sd = EFtune.curves;

   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to current curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }
   for (jj = 0; jj < 1000; jj++)
      sd->yval[ii][jj] = sd->yval[0][jj];

   tune_spc_moved[ii] = 1;
   tune_update_image();                                                    //  update image
   return;
}


//  Update image based on latest settings of all dialog controls.

void tune_update_image()
{
   int      px, py, ww, hh;

   if (sa_stat == 3 && sa_Npixel < 200000)                                 //  if relatively small area,
   {                                                                       //    process in main() thread
      for (py = sa_miny; py < sa_maxy; py++)
      for (px = sa_minx; px < sa_maxx; px++)
         tune_update_pixel(px,py);

      px = sa_minx;                                                        //  direct window update
      py = sa_miny;
      ww = sa_maxx - sa_minx;
      hh = sa_maxy - sa_miny;
      mwpaint3(px,py,ww,hh);
      CEF->Fmods++; 
      tune_updated++;
      return;
   }
   
   signal_thread();                                                        //  else use thread process
   return;
}


//  Update image based on latest settings of all dialog controls.

void * tune_thread(void *)
{
   void * tune_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(tune_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++; 
      tune_updated++;
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * tune_wthread(void *arg)                                             //  worker thread function
{
   int      px, py;
   int      index = *((int *) arg);
   
   for (py = index; py < E1pxm->hh; py += Nwt)
   for (px = 0; px < E1pxm->ww; px++)
      tune_update_pixel(px,py);

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


void tune_update_pixel(int px, int py)                                     //  process one pixel
{
   int         ii, kk, dist = 0;
   float       *pix1, *pix3;
   float       red1, green1, blue1, red3, green3, blue3;
   float       xval, brmean, brout, brightness;
   spldat      *sd = EFtune.curves;

   if (sa_stat == 3) {                                                     //  select area active
      ii = py * E1pxm->ww + px;
      dist = sa_pixmap[ii];                                                //  distance from edge
      if (! dist) return;                                                  //  pixel outside area
   }

   pix1 = PXMpix(E1pxm,px,py);                                             //  input pixel
   pix3 = PXMpix(E3pxm,px,py);                                             //  output pixel
   
   red1 = red3 = pix1[0];                                                  //  input and output RGB values
   green1 = green3 = pix1[1];
   blue1 = blue3 = pix1[2];

   brightness = 0.25 * red3 + 0.65 * green3 + 0.10 * blue3;                //  perceived brightness
   xval = brightness / 256.0;                                              //  curve x-value, 0 to 0.999

/* ------------------------------------------------------------------------

      brightness adjustment

      brightness curve values:
           0 = dark
         0.5 = normal, unchanged
         1.0 = 200% brightness, clipped
*/

   if (tune_spc_moved[1])                                                  //  curve has been edited
   {
      kk = 1000 * xval;                                                    //  x-value 
      if (kk > 999) kk = 999;
      brout = sd->yval[1][kk];                                             //  y-value, 0 to 1.0
      brout = 4 * brout - 2;                                               //  F-stops, -2 to +2
      brout = tune_pow[int(1000*brout+5000.5)];                            //  0.25 to 4.0
      
      red3 = red3 * brout;                                                 //  apply to all colors
      green3 = green3 * brout;                                             //  remove limit check
      blue3 = blue3 * brout;                                               //  (result is better)
   }
      
/* ------------------------------------------------------------------------

      color saturation curve values:
           0 = no color saturation (gray scale)
         0.5 = normal (initial unmodified RGB)
         1.0 = max. color saturation

      0.5 >> 0:  move all RGB values to their mean: (R+G+B)/3
      0.5 >> 1.0:  increase RGB spread to 2x original level
*/

   if (tune_spc_moved[2])                                                  //  curve has been edited
   {
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[2][kk];                                             //  saturation factor, 0 - 1
      brout = 2.0 * brout - 1.0;                                           //  -1.0 .. +1.0
      brmean = 0.333 * (red3 + green3 + blue3);
      red3 = red3 + brout * (red3 - brmean);                               //  simplified 
      green3 = green3 + brout * (green3 - brmean);
      blue3 = blue3 + brout * (blue3 - brmean);
      if (red3 < 0) red3 = 0;
      if (green3 < 0) green3 = 0;
      if (blue3 < 0) blue3 = 0;
   }

/* ------------------------------------------------------------------------

      color balance curve values:
         0 = 0.5 * original color
         0.5 = unmodified
         1.0 = 1.5 * original color, clipped
*/

   if (tune_spc_moved[3]) {                                                //  curve has been edited
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[3][kk];                                             //  0 to 1.0
      brout = 4 * brout - 2;                                               //  F-stops, -2 to +2
      brout = tune_pow[int(1000*brout+5000.5)];                            //  0.25 to 4.0
      red3 = red3 * brout;
   }

   if (tune_spc_moved[4]) {
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[4][kk];
      brout = 4 * brout - 2;
      brout = tune_pow[int(1000*brout+5000.5)];
      green3 = green3 * brout;
   }

   if (tune_spc_moved[5]) {
      kk = 1000 * xval;
      if (kk > 999) kk = 999;
      brout = sd->yval[5][kk];
      brout = 4 * brout - 2;
      brout = tune_pow[int(1000*brout+5000.5)];
      blue3 = blue3 * brout;
   }

//  if working within a select area, blend changes over distance from edge
      
   float       dold, dnew;

   if (sa_stat == 3 && dist < sa_blend) {
      dnew = 1.0 * dist / sa_blend;
      dold = 1.0 - dnew;
      red3 = dnew * red3 + dold * red1;
      green3 = dnew * green3 + dold * green1;
      blue3 = dnew * blue3 + dold * blue1;
   }

//  prevent overflow and set output RGB values

   float       cmax;

   cmax = red3;                                                            //  detect overflow
   if (green3 > cmax) cmax = green3;
   if (blue3 > cmax) cmax = blue3;
   
   if (cmax > 255.9) {                                                     //  stop overflow
      red3 = red3 * 255.9 / cmax;
      green3 = green3 * 255.9 / cmax;
      blue3 = blue3 * 255.9 / cmax;
   }

   pix3[0] = red3;                                                         //  0.0 to 255.9
   pix3[1] = green3;
   pix3[2] = blue3;
   
   return;
}


/**************************************************************************/

//  adjust image gamma using classic gamma curve

editfunc    EFgamma;
int         gamma_spc;                                                     //  current spline curve
int         gamma_updated;

void m_gamma(GtkWidget *, cchar *)
{
   int    gamma_dialog_event(zdialog* zd, cchar *event);
   void   gamma_curvedit(int spc);
   void * gamma_thread(void *);

   F1_help_topic = "gamma_curve";

   EFgamma.funcname = "gamma";
   EFgamma.Fprev = 1;                                                      //  use preview
   EFgamma.Farea = 2;                                                      //  select area usable
   EFgamma.menufunc = m_gamma;                                             //  restart allowed                 v.13.04
   EFgamma.threadfunc = gamma_thread;
   if (! edit_setup(EFgamma)) return;                                      //  setup edit
   
/***
          _________________________________________
         |                                         |                       //  5 curves are maintained:
         |                                         |                       //  curve 0: current display curve
         |                                         |                       //        1: curve for all colors
         |           curve edit area               |                       //        2,3,4: red, green, blue
         |                                         |
         |                                         |
         |_________________________________________|

            (o) all  (o) red  (o) green  (o) blue                          //  colors: all, red, green, blue
            [ + ] [ - ] [ + - ] [ - + ]  [reset]
            Curve File: [Open] [Save]   [histogram]
                                   [Done]  [Cancel]    
***/

   EFgamma.zd = zdialog_new(ZTX("adjust image gamma"),mWin,Bdone,Bcancel,null);
   zdialog *zd = EFgamma.zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hbrgb","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","all","hbrgb",Ball,"space=5");
   zdialog_add_widget(zd,"radio","red","hbrgb",Bred,"space=5");
   zdialog_add_widget(zd,"radio","green","hbrgb",Bgreen,"space=5");
   zdialog_add_widget(zd,"radio","blue","hbrgb",Bblue,"space=5");
   zdialog_add_widget(zd,"hbox","hbbu","dialog",0,"space=3");
   zdialog_add_widget(zd,"button","b ++","hbbu","   + +   ","space=4");
   zdialog_add_widget(zd,"button","b --","hbbu"," ‒ ‒ ","space=4");
   zdialog_add_widget(zd,"button","b +-","hbbu","  + ‒ ","space=4");
   zdialog_add_widget(zd,"button","b -+","hbbu"," ‒ +  ","space=4");
   zdialog_add_widget(zd,"button","reset","hbbu",Breset,"space=4");
   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen);
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=3");
   zdialog_add_widget(zd,"button","hist","hbcf",Bhistogram,"space=15");
   
   GtkWidget *frame = zdialog_widget(zd,"frame");
   spldat *sd = splcurve_init(frame,gamma_curvedit);
   EFgamma.curves = sd;

   sd->Nscale = 1;                                                         //  diagonal fixed line for neutral curve
   sd->xscale[0][0] = 0.00;
   sd->yscale[0][0] = 0.00;
   sd->xscale[1][0] = 1.00;
   sd->yscale[1][0] = 1.00;

   for (int ii = 0; ii < 5; ii++)                                          //  loop curves 0-4
   {
      sd->nap[ii] = 4;                                                     //  initial curves are flat
      sd->vert[ii] = 0;
      sd->apx[ii][0] = sd->apy[ii][0] = 0.01;
      sd->apx[ii][1] = sd->apy[ii][1] = 0.33;
      sd->apx[ii][2] = sd->apy[ii][2] = 0.67;
      sd->apx[ii][3] = sd->apy[ii][3] = 0.99;
      splcurve_generate(sd,ii);
   }

   sd->Nspc = 1;                                                           //  only curve 0 is shown 
   gamma_spc = 1;                                                          //  current active curve: 1 (all)
   gamma_updated = 0;                                                      //  no update yet
   
   zdialog_stuff(zd,"all",1);                                              //  stuff default selection, all

   zdialog_resize(zd,260,350);
   zdialog_run(zd,gamma_dialog_event,"save");                              //  run dialog - parallel

   return;
}


//  dialog event and completion callback function

int gamma_dialog_event(zdialog *zd, cchar *event)
{
   spldat      *sd = EFgamma.curves;
   int         ii, jj, nn, Fupdate = 0;
   float       step, step2, px, py;
   spldat      sdtemp;
   
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();                                     //  [done]
      else edit_cancel();                                                  //  [cancel] or [x]
      return 1;
   }

   if (strEqu(event,"save")) {                                             //  from f_save()                   v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"restart"))                                            //  new image                       v.13.04
      if (gamma_updated) Fupdate++;

   if (strnEqu(event,"b ",2))                                              //  button [ + ] etc.               v.12.04
   {                                                                       //  move entire curve
      step = 0.025;                                                        //  1/40
      step2 = 2 * step;

      for (jj = 1; jj < sd->nap[0]-1; jj++) {                              //  shift anchor points but         v.12.04
         px = sd->apx[0][jj];                                              //    not end points of curve 0
         py = sd->apy[0][jj];
         if (strEqu(event,"b ++")) py += step;
         if (strEqu(event,"b --")) py -= step;
         if (strEqu(event,"b +-"))  py += step - step2 * px;
         if (strEqu(event,"b -+"))  py -= step - step2 * px;
         if (py > 1) py = 1;
         if (py < 0) py = 0;
         sd->apy[0][jj] = py;
      }

      Fupdate++;
   }

   if (strEqu(event,"reset")) {                                            //  [reset] button         v.12.04
      sd->nap[0] = 4;                                                      //  restore curve 0 to initial state
      sd->apx[0][0] = sd->apy[0][0] = 0.01;
      sd->apx[0][1] = sd->apy[0][1] = 0.33;
      sd->apx[0][2] = sd->apy[0][2] = 0.67;
      sd->apx[0][3] = sd->apy[0][3] = 0.99;
      edit_reset();                                                        //  reset image mods          v.13.04
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      gamma_updated = 0;
   }

   if (strEqu(event,"blendwidth")) signal_thread();                        //  adjust area edge blending

   if (strEqu(event,"hist")) m_histogram(0,0);                             //  show brightness distribution

   if (strstr("all red green blue",event)) {                               //  new choice of curve
      zdialog_fetch(zd,event,ii);
      if (! ii) return 0;                                                  //  ignore button OFF events

      ii = strcmpv(event,"all","red","green","blue",null);
      gamma_spc = ii;                                                      //  1, 2, 3, 4

      sd->nap[0] = sd->nap[ii];                                            //  copy active curve ii to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"load"))                                               //  load 4 saved curves from file
   {
      sdtemp.Nspc = 4;
      sdtemp.drawarea = 0;
      nn = splcurve_load(&sdtemp);
      if (nn != 1) return 0;

      for (ii = 0; ii < 4; ii++) {                                         //  load curves 0-3 to curves 1-4
         sd->vert[ii+1] = sdtemp.vert[ii];
         sd->nap[ii+1] = sdtemp.nap[ii];
         for (jj = 0; jj < sdtemp.nap[ii]; jj++) {
            sd->apx[ii+1][jj] = sdtemp.apx[ii][jj];
            sd->apy[ii+1][jj] = sdtemp.apy[ii][jj];
         }
         splcurve_generate(sd,ii+1);
      }

      ii = gamma_spc;                                                      //  current active curve, 1-4

      sd->nap[0] = sd->nap[ii];                                            //  copy active curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"save"))                                               //  save 4 curves to a file
   {
      sdtemp.Nspc = 4;
      for (ii = 0; ii < 4; ii++) {
         sdtemp.vert[ii] = sd->vert[ii+1];
         sdtemp.nap[ii] = sd->nap[ii+1];
         for (jj = 0; jj < sd->nap[ii+1]; jj++) {
            sdtemp.apx[ii][jj] = sd->apx[ii+1][jj];
            sdtemp.apy[ii][jj] = sd->apy[ii+1][jj];
         }
      }

      splcurve_save(&sdtemp);
   }

   if (Fupdate)                                                            //  curve has changed
   {
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      
      ii = gamma_spc;                                                      //  active curve 1-4

      sd->nap[ii] = sd->nap[0];                                            //  copy curve 0 to active curve
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[ii][jj] = sd->apx[0][jj];
         sd->apy[ii][jj] = sd->apy[0][jj];
      }

      for (jj = 0; jj < 1000; jj++) 
         sd->yval[ii][jj] = sd->yval[0][jj];

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve

      if (gamma_spc == 1) {                                                //  "all" curve (1) was edited
         for (ii = 2; ii < 5; ii++) {                                      //  fix R/G/B curves (2-4) to match
            sd->nap[ii] = sd->nap[1];                                      //  v.12.04
            for (jj = 0; jj < sd->nap[1]; jj++) {
               sd->apx[ii][jj] = sd->apx[1][jj];
               sd->apy[ii][jj] = sd->apy[1][jj];
            }

            for (jj = 0; jj < 1000; jj++)
               sd->yval[ii][jj] = sd->yval[1][jj];
         }
      }

      signal_thread();                                                     //  trigger image update
   }

   return 1;
}


//  this function is called when a curve is edited

void gamma_curvedit(int spc)
{
   int      ii = gamma_spc, jj;
   spldat   *sd = EFgamma.curves;                                          //  active curve, 1-4

   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to active curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }

   for (jj = 0; jj < 1000; jj++) 
      sd->yval[ii][jj] = sd->yval[0][jj];

   if (gamma_spc == 1) {                                                   //  "all" curve (1) was edited
      for (ii = 2; ii < 5; ii++) {                                         //  fix R/G/B curves (2-4) to match
         sd->nap[ii] = sd->nap[1];
         for (jj = 0; jj < sd->nap[1]; jj++) {
            sd->apx[ii][jj] = sd->apx[1][jj];
            sd->apy[ii][jj] = sd->apy[1][jj];
         }

         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[1][jj];
      }
   }

   signal_thread();
   return;
}


//  gamma thread function

void * gamma_thread(void *arg)
{
   void * gamma_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(gamma_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      mwpaint2();                                                          //  update window
      CEF->Fmods++;                                                        //  image3 modified
      gamma_updated++;
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * gamma_wthread(void *arg)                                            //  worker thread function
{
   int         index = *((int *) arg);
   int         ii, dist = 0, px3, py3;
   float       *pix1, *pix3;
   float       red1, green1, blue1, maxrgb;
   float       red3, green3, blue3;
   float       coeff = 1000.0 / 256.0;
   float       dold, dnew;
   spldat      *sd = EFgamma.curves;

   for (py3 = index; py3 < E3pxm->hh; py3 += Nwt)                          //  loop output pixels
   for (px3 = 0; px3 < E3pxm->ww; px3++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py3 * E3pxm->ww + px3;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }
      
      pix1 = PXMpix(E1pxm,px3,py3);                                        //  input pixel
      pix3 = PXMpix(E3pxm,px3,py3);                                        //  output pixel
      
      red1 = pix1[0];                                                      //  input RGB brightness, 0-255.9
      green1 = pix1[1];
      blue1 = pix1[2];
      
      ii = coeff * red1;                                                   //  range 0-999 for gamma curve index
      red3 = 255.9 * sd->yval[2][ii];                                      //  output RGB brightness, 0-255.9
      ii = coeff * green1;
      green3 = 255.9 * sd->yval[3][ii];
      ii = coeff * blue1;
      blue3 = 255.9 * sd->yval[4][ii];
      
      maxrgb = red3;
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;
      
      if (maxrgb > 255.9) {                                                //  stop overflow
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
      }

      if (sa_stat == 3 && dist < sa_blend) {                               //  blend changes over blendwidth
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************

   Image Tone Mapping function
   enhance local contrast as opposed to overall contrast

   methodology:
   get brightness gradients for each pixel in 4 directions: SE SW NE NW
   amplify gradients using the edit curve (x-axis range 0-max. gradient)
   integrate 4 new brightness surfaces from the amplified gradients:
     - pixel brightness = prior pixel brightness + amplified gradient
     - the Amplify control varies amplification from zero to overamplified
   new pixel brightness = average from 4 calculated brightness surfaces

***************************************************************************/

float    *Tmap_brmap1;
float    *Tmap_brmap3[4];
int      Tmap_contrast99;
float    Tmap_amplify;

editfunc    EFtonemap;

void     Tmap_initz(zdialog *zd);
int      Tmap_dialog_event(zdialog *zd, cchar *event);
void     Tmap_curvedit(int);
void *   Tmap_thread(void *);


void m_tonemap(GtkWidget *, cchar *)
{
   F1_help_topic = "tone_mapping";

   cchar       *title = ZTX("Tone Mapping");

   EFtonemap.funcname = "tonemap";
   EFtonemap.Farea = 2;                                                    //  select area usable
   EFtonemap.menufunc = m_tonemap;                                         //  restart allowed                 v.13.04
   EFtonemap.threadfunc = Tmap_thread;
   if (! edit_setup(EFtonemap)) return;                                    //  setup: no preview, select area OK

/***
            _____________________________
           |                             |
           |                             |
           |    curve drawing area       |
           |                             |
           |_____________________________| 
            low       contrast       high

            Amplify ========[]=========

            Curve File: [ Open ] [ Save ]
***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFtonemap.zd = zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcL","hb1",ZTX("low"),"space=4");
   zdialog_add_widget(zd,"label","labcM","hb1",ZTX("contrast"),"expand");
   zdialog_add_widget(zd,"label","labcH","hb1",ZTX("high"),"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcon","hb2",ZTX("Amplify"),"space=5");
   zdialog_add_widget(zd,"hscale","amplify","hb2","0|1|0.005|0","expand");
   zdialog_add_widget(zd,"label","ampval","hb2",0,"space=5");

   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=8");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen,"space=5");
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=5");

   GtkWidget *frame = zdialog_widget(zd,"frame");                          //  set up curve edit
   spldat *sd = splcurve_init(frame,Tmap_curvedit);
   EFtonemap.curves = sd;

   sd->Nspc = 1;
   sd->vert[0] = 0;
   sd->nap[0] = 3;                                                         //  initial curve anchor points
   sd->apx[0][0] = 0.01;
   sd->apy[0][0] = 0.2;
   sd->apx[0][1] = 0.50;
   sd->apy[0][1] = 0.4;
   sd->apx[0][2] = 0.99;
   sd->apy[0][2] = 0.0;

   splcurve_generate(sd,0);                                                //  generate curve data

   Tmap_initz(zd);                                                         //  initialize                   v.13.04
   return;
}


//  initialization for new image file                                      //  v.13.04

void Tmap_initz(zdialog *zd)
{
   int         ii, cc, px, py;
   float       *pix1;
   int         jj, sum, limit, condist[100];
   
   if (Tmap_brmap1) {
      zfree(Tmap_brmap1);                                                  //  free memory
      for (int ii = 0; ii < 4; ii++)
      zfree(Tmap_brmap3[ii]);
   }

   cc = E1pxm->ww * E1pxm->hh * sizeof(float);                             //  allocate brightness map memory
   Tmap_brmap1 = (float *) zmalloc(cc,"tonemap");
   for (ii = 0; ii < 4; ii++)
      Tmap_brmap3[ii] = (float *) zmalloc(cc,"tonemap");

   for (py = 0; py < E1pxm->hh; py++)                                      //  map initial image brightness
   for (px = 0; px < E1pxm->ww; px++)
   {
      ii = py * E1pxm->ww + px;
      pix1 = PXMpix(E1pxm,px,py);
      Tmap_brmap1[ii] = pixbright(pix1);                                   //  pixel brightness 0-255.9
   }

   for (ii = 0; ii < 100; ii++) 
      condist[ii] = 0;

   for (py = 1; py < E1pxm->hh; py++)                                      //  map contrast distribution
   for (px = 1; px < E1pxm->ww; px++)
   {
      ii = py * E1pxm->ww + px;
      jj = 0.388 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-1]);             //  horiz. contrast, ranged 0 - 99
      condist[jj]++;
      jj = 0.388 * fabsf(Tmap_brmap1[ii] - Tmap_brmap1[ii-E1pxm->ww]);     //  vertical
      condist[jj]++;
   }
   
   sum = 0;
   limit = 0.99 * 2 * (E1pxm->ww-1) * (E1pxm->hh-1);                       //  find 99th percentile contrast

   for (ii = 0; ii < 100; ii++) {
      sum += condist[ii];
      if (sum > limit) break;
   }
   
   Tmap_contrast99 = 255.0 * ii / 100.0;                                   //  0 to 255
   if (Tmap_contrast99 < 4) Tmap_contrast99 = 4;                           //  rescale low-contrast image

   zdialog_resize(zd,300,300);
   zdialog_run(zd,Tmap_dialog_event,"save");                               //  run dialog - parallel

   Tmap_amplify = 0;
   return;
}


//  dialog event and completion callback function

int Tmap_dialog_event(zdialog *zd, cchar *event)
{
   spldat   *sd = EFtonemap.curves;
   char     text[8];
   
   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      zfree(Tmap_brmap1);                                                  //  free memory
      for (int ii = 0; ii < 4; ii++)
      zfree(Tmap_brmap3[ii]);
      Tmap_brmap1 = 0;                                                     //  v.13.04
      return 1;
   }
   
   if (strEqu(event,"restart")) {                                          //  new image                 v.13.04
      Tmap_initz(zd);
      zdialog_fetch(zd,"amplify",Tmap_amplify);
      if (Tmap_amplify) event = "amplify";
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"amplify")) {                                          //  slider value
      zdialog_fetch(zd,"amplify",Tmap_amplify);
      sprintf(text,"%.2f",Tmap_amplify);                                   //  numeric feedback
      zdialog_stuff(zd,"ampval",text);
      signal_thread();                                                     //  trigger update thread
   }

   if (strEqu(event,"load")) {                                             //  load saved curve 
      splcurve_load(sd);
      signal_thread();
      return 0;
   }

   if (strEqu(event,"save")) {                                             //  save curve to file 
      splcurve_save(sd);
      return 0;
   }
   
   return 0;
}


//  this function is called when the curve is edited

void Tmap_curvedit(int)
{
   signal_thread();
   return;
}


//  thread function

void * Tmap_thread(void *)
{
   void * Tmap_wthread1(void *arg);
   void * Tmap_wthread2(void *arg);

   int      ii;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (ii = 0; ii < 4; ii++)                                           //  start working threads1 (must be 4)
         start_wthread(Tmap_wthread1,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      for (ii = 0; ii < Nwt; ii++)                                         //  start working threads2
         start_wthread(Tmap_wthread2,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      CEF->Fmods++; 

      if (Tmap_amplify == 0)                                               //  v.13.04
         CEF->Fmods = 0;

      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  working threads

void * Tmap_wthread1(void *arg)
{
   int         ii, kk, bii, pii, dist = 0;
   int         px, px1, px2, pxinc;
   int         py, py1, py2, pyinc;
   float       b1, b3, xval, yval, grad;
   float       amplify, contrast99;
   spldat      *sd = EFtonemap.curves;

   contrast99 = Tmap_contrast99;                                           //  99th percentile contrast
   contrast99 = 1.0 / contrast99;                                          //  inverted

   amplify = pow(Tmap_amplify,0.4);                                        //  get amplification, 0 to 1    v.12.06

   bii = *((int *) arg);

   if (bii == 0) {                                                         //  direction SE
      px1 = 1; px2 = E3pxm->ww; pxinc = 1;
      py1 = 1; py2 = E3pxm->hh; pyinc = 1;
      pii = - 1 - E3pxm->ww;
   }
   
   else if (bii == 1) {                                                    //  direction SW
      px1 = E3pxm->ww-2; px2 = 0; pxinc = -1;
      py1 = 1; py2 = E3pxm->hh; pyinc = 1;
      pii = + 1 - E3pxm->ww;
   }
   
   else if (bii == 2) {                                                    //  direction NE
      px1 = 1; px2 = E3pxm->ww; pxinc = 1;
      py1 = E3pxm->hh-2; py2 = 0; pyinc = -1;
      pii = - 1 + E3pxm->ww;
   }
   
   else {   /* bii == 3 */                                                 //  direction NW
      px1 = E3pxm->ww-2; px2 = 0; pxinc = -1;
      py1 = E3pxm->hh-2; py2 = 0; pyinc = -1;
      pii = + 1 + E3pxm->ww;
   }

   for (ii = 0; ii < E3pxm->ww * E3pxm->hh; ii++)                          //  initial brightness map
      Tmap_brmap3[bii][ii] = Tmap_brmap1[ii];

   for (py = py1; py != py2; py += pyinc)                                  //  loop all image pixels
   for (px = px1; px != px2; px += pxinc)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside pixel
      }

      b1 = Tmap_brmap1[ii];                                                //  this pixel brightness
      grad = b1 - Tmap_brmap1[ii+pii];                                     //   - prior pixel --> gradient

      xval = fabsf(grad) * contrast99;                                     //  gradient scaled 0 to 1+
      kk = 1000.0 * xval;
      if (kk > 999) kk = 999;
      yval = 1.0 + 5.0 * sd->yval[0][kk];

      grad = grad * yval;                                                  //  magnified gradient

      b3 = Tmap_brmap3[bii][ii+pii] + grad;                                //  pixel brightness = prior + gradient
      b3 = (1.0 - amplify) * b1 + amplify * b3;                            //  constrain: push b3 toward b1

      if (b3 > 255) b3 = 255;                                              //  constrain
      if (b3 < 1) b3 = 1;

      Tmap_brmap3[bii][ii] = b3;                                           //  new pixel brightness
   }
   
   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


void * Tmap_wthread2(void *arg)
{
   float       *pix1, *pix3;
   int         index, ii, px, py, dist = 0;
   float       b1, b3, bf, f1, f2;
   float       red1, green1, blue1, red3, green3, blue3;

   index = *((int *) arg);

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active 
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      b1 = Tmap_brmap1[ii];                                                //  initial pixel brightness
      b3 = Tmap_brmap3[0][ii] + Tmap_brmap3[1][ii]                         //  new brightness = average of four
         + Tmap_brmap3[2][ii] + Tmap_brmap3[3][ii];                        //    calculated brightness surfaces

      bf = 0.25 * b3 / (b1 + 1);                                           //  brightness ratio
      
      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = bf * red1;                                                    //  output RGB
      if (red3 > 255.9) red3 = 255.9;
      green3 = bf * green1;
      if (green3 > 255.9) green3 = 255.9;
      blue3 = bf * blue1;
      if (blue3 > 255.9) blue3 = 255.9;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  flatten brightness distribution

float    flatten_value = 0;                                                //  flatten value, 0 - 100%
float    flatten_deband = 0;                                               //  flatten deband, 0 - 100%
float    flatten_brdist[256];

void  flatten_update_image();
void  flatten_distribution();                                              //  compute brightness distribution
void  flatten_pixel(int px, int py);                                       //  compute new pixel brightness

editfunc    EFflatten;

void m_flatten(GtkWidget *, cchar *)
{
   int    flatten_dialog_event(zdialog* zd, cchar *event);
   void * flatten_thread(void *);

   cchar  *title = ZTX("Flatten Brightness Distribution");
   F1_help_topic = "flatten";

   EFflatten.funcname = "flatten";
   EFflatten.Fprev = 1;                                                    //  preview
   EFflatten.Farea = 2;                                                    //  select area usable
   EFflatten.menufunc = m_flatten;                                         //  restart allowed              v.13.04
   EFflatten.threadfunc = flatten_thread;
   if (! edit_setup(EFflatten)) return;                                    //  setup edit

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFflatten.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");               //  flatten slider
   zdialog_add_widget(zd,"label","labfd","hb1",ZTX("Flatten"),"space=5");
   zdialog_add_widget(zd,"hscale","flatten","hb1","0|100|1|0","expand");
   zdialog_add_widget(zd,"label","fvalue","hb1",0,"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");               //  deband slider          v.12.06
   zdialog_add_widget(zd,"label","labdb","hb2",ZTX("Deband"),"space=5");
   zdialog_add_widget(zd,"hscale","deband","hb2","0|100|1|0","expand");
   zdialog_add_widget(zd,"label","dvalue","hb2",0,"space=5");

   zdialog_resize(zd,300,0);
   zdialog_run(zd,flatten_dialog_event,"save");                            //  run dialog - parallel
   
   flatten_value = flatten_deband = 0;
   return;
}


//  dialog event and completion callback function

int flatten_dialog_event(zdialog *zd, cchar *event)                        //  flatten dialog event function
{
   char     text[8];

   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 1;
   }
   
   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"restart"))                                            //  new image                    v.13.04
      if (flatten_value) flatten_update_image();
   
   if (strEqu(event,"blendwidth")) flatten_update_image();

   if (strEqu(event,"flatten")) {
      zdialog_fetch(zd,"flatten",flatten_value);                           //  get slider value
      sprintf(text,"%.0f",flatten_value);                                  //  numeric feedback 
      zdialog_stuff(zd,"fvalue",text);
      flatten_update_image();                                              //  trigger update thread
   }
   
   if (strEqu(event,"deband")) {
      zdialog_fetch(zd,"deband",flatten_deband);                           //  get slider value             v.12.06
      sprintf(text,"%.0f",flatten_deband);                                 //  numeric feedback 
      zdialog_stuff(zd,"dvalue",text);
      flatten_update_image();                                              //  trigger update thread
   }
   
   return 1;
}


//  Update image based on select area and dialog controls

void flatten_update_image()
{
   int      px, py, ww, hh;
   
   if (flatten_value == 0) {                                               //  v.13.04
      edit_reset();
      return;
   }

   if (sa_stat == 3 && sa_Npixel < 200000)                                 //  if relatively small area,
   {                                                                       //    process in main() thread
      flatten_distribution();

      for (py = sa_miny; py < sa_maxy; py++)
      for (px = sa_minx; px < sa_maxx; px++)
         flatten_pixel(px,py);

      px = sa_minx;                                                        //  direct window update
      py = sa_miny;
      ww = sa_maxx - sa_minx;
      hh = sa_maxy - sa_miny;
      mwpaint3(px,py,ww,hh); 
      CEF->Fmods++; 
   }
   
   else signal_thread();                                                   //  use thread process
   return;
}


//  thread function - use multiple working threads

void * flatten_thread(void *)
{
   void  * flatten_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      flatten_distribution();
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(flatten_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++; 
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * flatten_wthread(void *arg)                                          //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py;

   for (py = index; py < E1pxm->hh; py += Nwt)                             //  flatten brightness distribution
   for (px = 0; px < E1pxm->ww; px++)
      flatten_pixel(px,py);

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


//  compute brightness distribution for image or selected area

void flatten_distribution()
{
   int         px, py, ii, npix;
   int         ww = E1pxm->ww, hh = E1pxm->hh;
   float       bright1;
   float       *pix1;

   for (ii = 0; ii < 256; ii++)                                            //  clear brightness distribution data
      flatten_brdist[ii] = 0;

   if (sa_stat == 3)                                                       //  process selected area
   {
      for (ii = npix = 0; ii < ww * hh; ii++)
      {
         if (! sa_pixmap[ii]) continue;
         py = ii / ww;
         px = ii - py * ww;
         pix1 = PXMpix(E1pxm,px,py);
         bright1 = pixbright(pix1);
         flatten_brdist[int(bright1)]++;
         npix++;
      }
      
      for (ii = 1; ii < 256; ii++)                                         //  cumulative brightness distribution
         flatten_brdist[ii] += flatten_brdist[ii-1];                       //   0 ... npix

      for (ii = 0; ii < 256; ii++)
         flatten_brdist[ii] = flatten_brdist[ii]                           //  multiplier per brightness level
                            / npix * 256.0 / (ii + 1);
   }

   else                                                                    //  process whole image
   {
      for (py = 0; py < hh; py++)                                          //  compute brightness distribution
      for (px = 0; px < ww; px++)
      {
         pix1 = PXMpix(E1pxm,px,py);
         bright1 = pixbright(pix1);
         flatten_brdist[int(bright1)]++;
      }
      
      for (ii = 1; ii < 256; ii++)                                         //  cumulative brightness distribution
         flatten_brdist[ii] += flatten_brdist[ii-1];                       //   0 ... (ww * hh)

      for (ii = 0; ii < 256; ii++)
         flatten_brdist[ii] = flatten_brdist[ii]                           //  multiplier per brightness level 
                            / (ww * hh) * 256.0 / (ii + 1);
   }
   
   return;
}


//  compute new brightness for one pixel

void flatten_pixel(int px, int py)
{
   int         ii, dist = 0;
   float       *pix1, *pix3;
   float       fold, fnew, dold, dnew, cmax;
   float       red1, green1, blue1, red3, green3, blue3;
   float       bright1, bright2, deband;

   if (sa_stat == 3) {                                                     //  select area active
      ii = py * E1pxm->ww + px;   
      dist = sa_pixmap[ii];                                                //  distance from edge
      if (! dist) return;                                                  //  outside pixel
   }

   pix1 = PXMpix(E1pxm,px,py);                                             //  input pixel
   pix3 = PXMpix(E3pxm,px,py);                                             //  output pixel
      
   red1 = pix1[0];
   green1 = pix1[1];
   blue1 = pix1[2];

   bright1 = 0.25 * red1 + 0.65 * green1 + 0.10 * blue1;                   //  input pixel brightness
   bright2 = flatten_brdist[int(bright1)];                                 //  output adjustment factor

   red3 = bright2 * red1;                                                  //  fully flattened brightness
   green3 = bright2 * green1;
   blue3 = bright2 * blue1;

   deband = (256.0 - bright1) / 256.0;                                     //  bright1 = 0 to 256  >>  deband = 1 to 0
   deband += (1.0 - deband) * (1.0 - 0.01 * flatten_deband);               //  deband = 1 to 1  >>  1 to 0

   fnew = 0.01 * flatten_value;                                            //  how much to flatten, 0 to 1
   fnew = deband * fnew;                                                   //  attenuate flatten of brighter pixels
   fold = 1.0 - fnew;                                                      //  how much to retain, 1 to 0

   red3 = fnew * red3 + fold * red1;                                       //  blend new and old brightness
   green3 = fnew * green3 + fold * green1;
   blue3 = fnew * blue3 + fold * blue1;

   if (sa_stat == 3 && dist < sa_blend) {                                  //  select area is active,
      dnew = 1.0 * dist / sa_blend;                                        //    blend changes over sa_blend
      dold = 1.0 - dnew;
      red3 = dnew * red3 + dold * red1;
      green3 = dnew * green3 + dold * green1;
      blue3 = dnew * blue3 + dold * blue1;
   }

   cmax = red3;                                                            //  stop overflow, keep color balance
   if (green3 > cmax) cmax = green3;
   if (blue3 > cmax) cmax = blue3;
   if (cmax > 255.9) {
      cmax = 255.9 / cmax;
      red3 = red3 * cmax;
      green3 = green3 * cmax;
      blue3 = blue3 * cmax;
   }

   pix3[0] = red3;
   pix3[1] = green3;
   pix3[2] = blue3;

   return;
}


/**************************************************************************/

//  Clip pixels outside the given low/high range and expand the remaining
//  brightness range back to the full range of 0 to 255.

float       xbrangeD, xbrangeB;
editfunc    EFxbrange;

void m_xbrange(GtkWidget *, cchar *) 
{
   int      xbrange_dialog_event(zdialog *zd, cchar *event);
   void *   xbrange_thread(void *);
   
   cchar  *title = ZTX("Expand Brightness Range");
   cchar  *labD = ZTX("dark pixels");
   cchar  *labB = ZTX("bright pixels");

   F1_help_topic = "expand_brightness";
   
   EFxbrange.funcname = "bright-range";
   EFxbrange.Fprev = 1;                                                    //  use preview
   EFxbrange.Farea = 2;                                                    //  select area usable
   EFxbrange.menufunc = m_xbrange;                                         //  restart allowed              v.13.04
   EFxbrange.threadfunc = xbrange_thread;                                  //  thread function
   if (! edit_setup(EFxbrange)) return;                                    //  setup edit

   m_histogram(0,0);                                                       //  show brightness distribution 

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFxbrange.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"expand");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|expand");
   zdialog_add_widget(zd,"vbox","vb3","hb1",0,"homog|space=8");
   zdialog_add_widget(zd,"label","labD","vb1",labD);
   zdialog_add_widget(zd,"label","labB","vb1",labB);
   zdialog_add_widget(zd,"hscale","clipD","vb2","0|99|0.5|0");
   zdialog_add_widget(zd,"hscale","clipB","vb2","0|99|0.5|0");
   zdialog_add_widget(zd,"label","darkval","vb3");
   zdialog_add_widget(zd,"label","briteval","vb3");

   zdialog_resize(zd,300,0);
   zdialog_run(zd,xbrange_dialog_event,"save");                            //  run dialog - parallel
   
   xbrangeD = xbrangeB = 0;                                                //  initial clip = 0
   return;
}


//  dialog event and completion callback function

int xbrange_dialog_event(zdialog *zd, cchar *event)
{
   char     text[8];

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }

   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"restart"))                                            //  new image                    v.13.04
      if (xbrangeD || xbrangeB) signal_thread();
   
   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"clipD")) {
      zdialog_fetch(zd,"clipD",xbrangeD);
      signal_thread();
   }

   if (strEqu(event,"clipB")) {
      zdialog_fetch(zd,"clipB",xbrangeB);
      signal_thread();
   }
   
   sprintf(text,"%.0f",xbrangeD);                                          //  numeric feedback
   zdialog_stuff(zd,"darkval",text);
   sprintf(text,"%.0f",xbrangeB);
   zdialog_stuff(zd,"briteval",text);

   return 0;
}


//  thread function

void * xbrange_thread(void *)
{
   int         ii, px, py, dist = 0;
   float       dark, bright, b1, b3, bf, f1, f2;
   float       red1, green1, blue1, red3, green3, blue3;
   float       *pix1, *pix3;
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      if (! xbrangeD && ! xbrangeB) {                                      //  v.13.04
         CEF->Fmods = 0;
         continue;
      }

      dark = 0.01 * xbrangeD * 256;                                        //  clipping brightness levels
      bright = (1.0 - 0.01 * xbrangeB) * 256;

      for (py = 0; py < E3pxm->hh; py++)                                   //  loop all image pixels
      for (px = 0; px < E3pxm->ww; px++)
      {
         ii = py * E3pxm->ww + px;

         if (sa_stat == 3) {                                               //  select area active
            dist = sa_pixmap[ii];                                          //  distance from edge
            if (! dist) continue;                                          //  pixel is outside area
         }

         pix1 = PXMpix(E1pxm,px,py);                                       //  input pixel
         pix3 = PXMpix(E3pxm,px,py);                                       //  output pixel

         b1 = pixbright(pix1);

         if (b1 < dark)                                                    //  clip dark pixels
            b3 = 0;
         else if (b1 > bright)                                             //  clip bright pixels
            b3 = bright;
         else b3 = b1;
         
         if (b3 > dark) b3 = b3 - dark;                                    //  expand the rest
         b3 = b3 * (255.0 / (bright - dark));

         bf = b3 / (b1 + 1);                                               //  brightness ratio
         
         red1 = pix1[0];                                                   //  input RGB 
         green1 = pix1[1];
         blue1 = pix1[2];
         
         red3 = bf * red1;                                                 //  output RGB
         green3 = bf * green1;
         blue3 = bf * blue1;
         
         if (red3 > 255) red3 = 255;
         if (green3 > 255) green3 = 255;
         if (blue3 > 255) blue3 = 255;
         
         if (sa_stat == 3 && dist < sa_blend) {                            //  select area is active,
            f1 = 1.0 * dist / sa_blend;                                    //    blend changes over sa_blend
            f2 = 1.0 - f1;
            red3 = f1 * red3 + f2 * red1;
            green3 = f1 * green3 + f2 * green1;
            blue3 = f1 * blue3 + f2 * blue1;
         }
         
         pix3[0] = red3;
         pix3[1] = green3;
         pix3[2] = blue3;
      }

      CEF->Fmods++; 
      mwpaint2();
   }

   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  adjust white balance

void   whitebal_mousefunc();

float       whitebal_red, whitebal_green, whitebal_blue;
float       whitebal_reduce;
editfunc    EFwhitebal;

void m_whitebal(GtkWidget *, cchar *)
{
   int    whitebal_dialog_event(zdialog *zd, cchar *event);
   void * whitebal_thread(void *);

   cchar  *wbtitle = ZTX("Adjust White Balance");
   cchar  *wbhelp = ZTX("Click white or gray image location");

   F1_help_topic = "white_balance";
   
   EFwhitebal.funcname = "white-balance";
   EFwhitebal.Fprev = 1;                                                   //  use preview
   EFwhitebal.Farea = 2;                                                   //  select area usable
   EFwhitebal.menufunc = m_whitebal;                                       //  restart allowed              v.13.04
   EFwhitebal.threadfunc = whitebal_thread;                                //  thread function
   EFwhitebal.mousefunc = whitebal_mousefunc;                              //  mouse function

   if (! edit_setup(EFwhitebal)) return;                                   //  setup edit: preview

   zdialog *zd = zdialog_new(wbtitle,mWin,Bdone,Bcancel,null);
   EFwhitebal.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=10");
   zdialog_add_widget(zd,"label","labwbh","hb1",wbhelp,"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labpix","hb2","pixel:","space=3");
   zdialog_add_widget(zd,"label","pixel","hb2","0000 0000");
   zdialog_add_widget(zd,"label","labrgb","hb2","   RGB:");
   zdialog_add_widget(zd,"label","rgb","hb2","000 000 000");
   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labdamp","hb3",ZTX("reduce"),"space=3");
   zdialog_add_widget(zd,"hscale","reduce","hb3","0|1|0.01|0","expand");

   zdialog_run(zd,whitebal_dialog_event,"save");                           //  run dialog - parallel

   whitebal_red = whitebal_green = whitebal_blue = 1.0;
   whitebal_reduce = 0.0;

   takeMouse(whitebal_mousefunc,dragcursor);                               //  connect mouse function
   return;
}


//  dialog event and completion callback function

int whitebal_dialog_event(zdialog *zd, cchar *event)                       //  dialog event function
{
   if (zd->zstat)
   {
      if (zd->zstat == 1) edit_done();                                     //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 0;
   }
   
   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(whitebal_mousefunc,dragcursor);                            //  connect mouse function
   
   if (strEqu(event,"reduce")) {                                           //  reduce impact       v.12.04
      zdialog_fetch(zd,"reduce",whitebal_reduce);
      signal_thread();
   }
   
   if (strEqu(event,"blendwidth")) 
      signal_thread();
   
   return 1;
}


//  get nominal white color from mouse click position

void whitebal_mousefunc()                                                  //  mouse function
{
   int         px, py, dx, dy;
   float       red, green, blue, rgbmean;
   char        work[40];
   float       *ppix;
   zdialog     *zd = EFwhitebal.zd;
   
   if (! LMclick) return;
   
   LMclick = 0;
   px = Mxclick;                                                           //  mouse click position
   py = Myclick;
   
   if (px < 2) px = 2;                                                     //  pull back from edge
   if (px > E3pxm->ww-3) px = E3pxm->ww-3;
   if (py < 2) py = 2;
   if (py > E3pxm->hh-3) py = E3pxm->hh-3;
   
   red = green = blue = 0;

   for (dy = -2; dy <= 2; dy++)                                            //  5x5 block around mouse position
   for (dx = -2; dx <= 2; dx++)
   {
      ppix = PXMpix(E1pxm,px+dx,py+dy);                                    //  input image
      red += ppix[0];
      green += ppix[1];
      blue += ppix[2];
   }

   red = red / 25.0;                                                       //  mean RGB levels 
   green = green / 25.0;
   blue = blue / 25.0;
   rgbmean = (red + green + blue) / 3.0;
   
   whitebal_red = rgbmean / red;                                           //  = 1.0 if all RGB are equal
   whitebal_green = rgbmean / green;
   whitebal_blue = rgbmean / blue;

   snprintf(work,40,"%d %d",px,py);
   zdialog_stuff(zd,"pixel",work);

   snprintf(work,40,"%7.3f %7.3f %7.3f",red/256,green/256,blue/256);
   zdialog_stuff(zd,"rgb",work);
   
   signal_thread();                                                        //  trigger image update
   return;
}


//  Update image based on neutral pixel that was clicked

void * whitebal_thread(void *)
{
   void * whitebal_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(whitebal_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++; 
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * whitebal_wthread(void *arg)                                         //  worker thread function
{
   int         index = *((int *) arg);
   int         px, py, ii, dist = 0;
   float       *pix1, *pix3;
   float       red1, green1, blue1;
   float       red3, green3, blue3;
   float       brmax, dold, dnew;
   float       Fred, Fgreen, Fblue;

   Fred = whitebal_red + (1.0 - whitebal_red) * whitebal_reduce;           //  as reduce goes from 0 to 1.0.
   Fgreen = whitebal_green + (1.0 - whitebal_green) * whitebal_reduce;     //    Fcolors go from initial to 1.0
   Fblue = whitebal_blue + (1.0 - whitebal_blue) * whitebal_reduce;        //  12.04

   for (py = index; py < E1pxm->hh; py += Nwt)
   for (px = 0; px < E1pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E1pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
      }

      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel
      
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = Fred * red1;                                                  //  change color ratios
      green3 = Fgreen * green1;
      blue3 = Fblue * blue1;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         dnew = 1.0 * dist / sa_blend;                                     //    blend changes over sa_blend
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }
      
      brmax = red3;                                                        //  brmax = brightest color
      if (green3 > brmax) brmax = green3;
      if (blue3 > brmax) brmax = blue3;
      
      if (brmax > 255.9) {                                                 //  if overflow, reduce
         brmax = 255.9 / brmax;
         red3 = red3 * brmax;
         green3 = green3 * brmax;
         blue3 = blue3 * brmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }
   
   exit_wthread();
   return 0;                                                               //  not executed, stop gcc warning
}


/**************************************************************************/

//  ramp brightness across image, vertical and horizontal gradients

editfunc    EFbrramp;
int         brramp_spc;                                                    //  current spline curves (2 at a time)

void m_brightramp(GtkWidget *, cchar *)                                    //  upgrade for 4 colors
{
   int    brramp_dialog_event(zdialog* zd, cchar *event);
   void   brramp_curvedit(int spc);
   void * brramp_thread(void *);
   void   brramp_init();

   F1_help_topic = "brightness_ramp";

   EFbrramp.funcname = "bright-ramp";
   EFbrramp.Fprev = 1;                                                     //  use preview
   EFbrramp.Farea = 2;                                                     //  select area usable
   EFbrramp.menufunc = m_brightramp;                                       //  allow restart                   v.13.04
   EFbrramp.threadfunc = brramp_thread;
   if (! edit_setup(EFbrramp)) return;                                     //  setup edit
   
/***
          _________________________________________
         |                                         |
         |                                         |
         |           curve edit area               |
         |                                         |
         |_________________________________________|

            (o) all  (o) red  (o) green  (o) blue                          //  colors: all, red, green, blue
            Curve File: [Open]  [Save]                                     //  each has vert. and horiz. curve
                                   [Done]  [Cancel]    

***/

   zdialog *zd = zdialog_new(ZTX("Ramp brightness across image"),mWin,Bdone,Bcancel,null);
   EFbrramp.zd = zd;

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5|expand");
   zdialog_add_widget(zd,"vbox","vb1","hb1");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"expand");
   zdialog_add_widget(zd,"label","lmax","vb1","+","space=3");
   zdialog_add_widget(zd,"label","lspace","vb1",0,"expand");
   zdialog_add_widget(zd,"label","lmin","vb1","‒","space=3");
   zdialog_add_widget(zd,"label","lspace","vb1");
   zdialog_add_widget(zd,"frame","frame","vb2",0,"expand");

   zdialog_add_widget(zd,"hbox","hb2","vb2");
   zdialog_add_widget(zd,"label","lmin","hb2","‒","space=3");
   zdialog_add_widget(zd,"label","lspace","hb2",0,"expand");
   zdialog_add_widget(zd,"label","lmax","hb2","+","space=3");
   zdialog_add_widget(zd,"hbox","hbrgb","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","all","hbrgb",Ball,"space=5");
   zdialog_add_widget(zd,"radio","red","hbrgb",Bred,"space=5");
   zdialog_add_widget(zd,"radio","green","hbrgb",Bgreen,"space=5");
   zdialog_add_widget(zd,"radio","blue","hbrgb",Bblue,"space=5");
   zdialog_add_widget(zd,"hbox","hbcf","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcf","hbcf",Bcurvefile,"space=8");
   zdialog_add_widget(zd,"button","load","hbcf",Bopen,"space=5");
   zdialog_add_widget(zd,"button","save","hbcf",Bsave,"space=5");

   GtkWidget *frame = zdialog_widget(zd,"frame");
   spldat *sd = splcurve_init(frame,brramp_curvedit);
   EFbrramp.curves = sd;
   
   brramp_init();                                                          //  set up initial flat curves

   sd->Nspc = 2;                                                           //  only curves 0-1 are shown
   brramp_spc = 2;                                                         //  current curves: all (2-3)
   
   zdialog_stuff(zd,"all",1);                                              //  stuff default selection, all

   zdialog_resize(zd,260,350);
   zdialog_run(zd,brramp_dialog_event,"save");                             //  run dialog - parallel
   return;
}


//  initialize all curves to flat

void brramp_init()
{
   spldat *sd = EFbrramp.curves;

   for (int ii = 0; ii < 10; ii++)                                         //  loop curves: current (0-1), all (2-3),
   {                                                                       //    red (4-5), green (6-7), blue (8-9)
      sd->nap[ii] = 2;                                                     //  horizontal curve ii
      sd->vert[ii] = 0;
      sd->apx[ii][0] = 0.01;
      sd->apx[ii][1] = 0.99;
      sd->apy[ii][0] = sd->apy[ii][1] = 0.5;
      splcurve_generate(sd,ii);

      ii++;
      sd->nap[ii] = 2;                                                     //  vertical curve ii+1
      sd->vert[ii] = 1;
      sd->apx[ii][0] = 0.01;
      sd->apx[ii][1] = 0.99;
      sd->apy[ii][0] = sd->apy[ii][1] = 0.5;
      splcurve_generate(sd,ii);
   }
   
   return;
}


//  dialog event and completion callback function

int brramp_dialog_event(zdialog *zd, cchar *event)
{
   spldat      *sd = EFbrramp.curves;
   int         ii, jj, nn, Fupdate = 0;
   spldat      sdtemp;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) edit_done();
      else edit_cancel();
      return 1;
   }

   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"blendwidth")) signal_thread();

   if (strstr("all red green blue",event)) {                               //  new choice of curve
      zdialog_fetch(zd,event,ii);
      if (! ii) return 0;                                                  //  ignore button OFF events

      ii = strcmpv(event,"all","red","green","blue",null);
      brramp_spc = ii = 2 * ii;                                            //  2, 4, 6, 8

      sd->nap[0] = sd->nap[ii];                                            //  copy active curves ii/ii+1 to curves 0/1
      sd->nap[1] = sd->nap[ii+1];
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[1][jj] = sd->apx[ii+1][jj];
         sd->apy[1][jj] = sd->apy[ii+1][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"load"))                                               //  load 8 saved curves from file
   {
      sdtemp.Nspc = 8;
      sdtemp.drawarea = 0;
      nn = splcurve_load(&sdtemp);
      if (nn != 1) return 0;

      for (ii = 0; ii < 8; ii++) {                                         //  load curves 0-7 to curves 2-9
         sd->vert[ii+2] = sdtemp.vert[ii];
         sd->nap[ii+2] = sdtemp.nap[ii];
         for (jj = 0; jj < sdtemp.nap[ii]; jj++) {
            sd->apx[ii+2][jj] = sdtemp.apx[ii][jj];
            sd->apy[ii+2][jj] = sdtemp.apy[ii][jj];
         }
         splcurve_generate(sd,ii+2);
      }

      ii = brramp_spc;                                                     //  current active curves: 2, 4, 6, 8

      sd->nap[0] = sd->nap[ii];                                            //  copy 2 active curves to curves 0-1
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }

      sd->nap[1] = sd->nap[ii+1];
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[1][jj] = sd->apx[ii+1][jj];
         sd->apy[1][jj] = sd->apy[ii+1][jj];
      }

      Fupdate++;
   }

   if (strEqu(event,"save"))                                               //  save 8 curves to file
   {
      sdtemp.Nspc = 8;
      for (ii = 0; ii < 8; ii++) {
         sdtemp.vert[ii] = sd->vert[ii+2];
         sdtemp.nap[ii] = sd->nap[ii+2];
         for (jj = 0; jj < sd->nap[ii+2]; jj++) {
            sdtemp.apx[ii][jj] = sd->apx[ii+2][jj];
            sdtemp.apy[ii][jj] = sd->apy[ii+2][jj];
         }
      }

      splcurve_save(&sdtemp);
   }

   if (Fupdate)                                                            //  curve has changed
   {
      splcurve_generate(sd,0);                                             //  regenerate curves 0-1
      splcurve_generate(sd,1);
      
      ii = brramp_spc;                                                     //  active curves

      sd->nap[ii] = sd->nap[0];                                            //  copy curves 0-1 to active curves
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[ii][jj] = sd->apx[0][jj];
         sd->apy[ii][jj] = sd->apy[0][jj];
      }

      sd->nap[ii+1] = sd->nap[1];
      for (jj = 0; jj < sd->nap[1]; jj++) {
         sd->apx[ii+1][jj] = sd->apx[1][jj];
         sd->apy[ii+1][jj] = sd->apy[1][jj];
      }

      for (jj = 0; jj < 1000; jj++) 
      {
         sd->yval[ii][jj] = sd->yval[0][jj];
         sd->yval[ii+1][jj] = sd->yval[1][jj];
      }

      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve

      signal_thread();                                                     //  trigger image update
   }

   return 1;
}


//  this function is called when a curve is edited

void brramp_curvedit(int spc)
{
   int      ii = brramp_spc, jj;
   spldat   *sd = EFbrramp.curves;                                         //  active curves: 2, 4, 6, 8 (and +1)

   sd->nap[ii] = sd->nap[0];                                               //  copy curves 0-1 to active curves
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }

   sd->nap[ii+1] = sd->nap[1];
   for (jj = 0; jj < sd->nap[1]; jj++) {
      sd->apx[ii+1][jj] = sd->apx[1][jj];
      sd->apy[ii+1][jj] = sd->apy[1][jj];
   }

   for (jj = 0; jj < 1000; jj++) {
      sd->yval[ii][jj] = sd->yval[0][jj];
      sd->yval[ii+1][jj] = sd->yval[1][jj];
   }

   if (brramp_spc == 2) {                                                  //  "all" curve was edited
      for (ii = 4; ii < 10; ii += 2) {                                     //  fix R/G/B curves to match
         sd->nap[ii] = sd->nap[2];
         for (jj = 0; jj < sd->nap[2]; jj++) {
            sd->apx[ii][jj] = sd->apx[2][jj];
            sd->apy[ii][jj] = sd->apy[2][jj];
         }
         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[2][jj];
      }
      for (ii = 5; ii < 10; ii += 2) {
         sd->nap[ii] = sd->nap[3];
         for (jj = 0; jj < sd->nap[3]; jj++) {
            sd->apx[ii][jj] = sd->apx[3][jj];
            sd->apy[ii][jj] = sd->apy[3][jj];
         }
         for (jj = 0; jj < 1000; jj++)
            sd->yval[ii][jj] = sd->yval[3][jj];
      }
   }

   signal_thread();
   return;
}


//  brramp thread function

void * brramp_thread(void *arg)
{
   void * brramp_wthread(void *);
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(brramp_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      mwpaint2();                                                          //  update window
      CEF->Fmods++;                                                        //  image3 modified
   }

   return 0;                                                               //  not executed, stop g++ warning
}


void * brramp_wthread(void *arg)                                           //  worker thread function
{
   int         index = *((int *) arg);
   int         ii, jj, kk, dist = 0, px3, py3;
   float       *pix1, *pix3;
   float       red1, green1, blue1, maxrgb;
   float       red3, green3, blue3;
   float       dispx, dispy;
   float       hramp[3], vramp[3], tramp[3];
   float       spanw, spanh, dold, dnew;
   spldat      *sd = EFbrramp.curves;

   if (sa_stat == 3) {                                                     //  if select area active, ramp
      spanw = sa_maxx - sa_minx;                                           //    brightness over enclosing rectangle
      spanh = sa_maxy - sa_miny;
   }
   else {
      spanw = E3pxm->ww;                                                   //  else over entire image
      spanh = E3pxm->hh;
   }   

   for (py3 = index; py3 < E3pxm->hh; py3 += Nwt)                          //  loop output pixels
   for (px3 = 0; px3 < E3pxm->ww; px3++)
   {
      if (sa_stat == 3) {                                                  //  select area active 
         ii = py3 * E3pxm->ww + px3;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel outside area
         dispx = (px3 - sa_minx) / spanw;
         dispy = (py3 - sa_miny) / spanh;
      }
      else {
         dispx = px3 / spanw;                                              //  left > right = 0 to 1
         dispy = py3 / spanh;                                              //  top > bottom = 0 to 1
      }
      
      kk = 1000 * dispx;                                                   //  conv. dispx 0-1 to 0-1000
      if (kk > 999) kk = 999;

      for (ii = 0, jj = 4; ii < 3; ii++, jj+=2)                            //  horz. curves for dispx and R/G/B
         hramp[ii] = sd->yval[jj][kk] - 0.5;                               //  scale to -0.5 to +0.5
      
      kk = 1000 * dispy;                                                   //  conv. dispy
      if (kk > 999) kk = 999;

      for (ii = 0, jj = 5; ii < 3; ii++, jj+=2)                            //  vert. curves
         vramp[ii] = sd->yval[jj][kk] - 0.5;

      for (ii = 0; ii < 3; ii++)                                           //  total R/G/B change
         tramp[ii] = 1.0 + hramp[ii] + vramp[ii];                          //  scale to 0.0 to 2.0

      pix1 = PXMpix(E1pxm,px3,py3);                                        //  input pixel
      red1 = pix1[0];
      green1 = pix1[1];
      blue1 = pix1[2];
      
      red3 = red1 * tramp[0];                                              //  change R/G/B
      green3 = green1 * tramp[1];
      blue3 = blue1 * tramp[2];

      maxrgb = red3;
      if (green3 > maxrgb) maxrgb = green3;
      if (blue3 > maxrgb) maxrgb = blue3;
      
      if (maxrgb > 255.9) {                                                //  stop overflow
         red3 = red3 * 255.9 / maxrgb;
         green3 = green3 * 255.9 / maxrgb;
         blue3 = blue3 * 255.9 / maxrgb;
      }

      if (sa_stat == 3 && dist < sa_blend) {                               //  blend changes over blendwidth
         dnew = 1.0 * dist / sa_blend;
         dold = 1.0 - dnew;
         red3 = dnew * red3 + dold * red1;
         green3 = dnew * green3 + dold * green1;
         blue3 = dnew * blue3 + dold * blue1;
      }

      pix3 = PXMpix(E3pxm,px3,py3);                                        //  output pixel
      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************

   Vignette function
   
   1. Change the brightness from center to edge using a curve.
   2. Change the color from center to edge using a color and a curve.
      (the pixel varies between original RGB and selected color)
   3. Mouse click or drag on image sets a new vignette center.

***************************************************************************/

void  vign_mousefunc();                                                    //  v.12.09

editfunc    EFvignette;
uint8       vignette_RGB[3] = { 0, 0, 255 };
int         vignette_spc;
float       vign_cx, vign_cy;
float       vign_rad;


void m_vignette(GtkWidget *, cchar *)
{
   int      Vign_dialog_event(zdialog *zd, cchar *event);
   void     Vign_curvedit(int);
   void *   Vign_thread(void *);

   F1_help_topic = "vignette";

   cchar       *title = ZTX("Vignette Tool");

   EFvignette.funcname = "vignette";
   EFvignette.Farea = 2;                                                   //  select area usable
   EFvignette.Fprev = 1;                                                   //  use preview image
   EFvignette.menufunc = m_vignette;                                       //  allow restart                   v.13.04
   EFvignette.threadfunc = Vign_thread;                                    //  thread function
   EFvignette.mousefunc = vign_mousefunc;                                  //  mouse function

   if (! edit_setup(EFvignette)) return;                                   //  setup edit

/***
            _____________________________
           |                             |
           |                             |
           |    curve drawing area       |
           |                             |
           |_____________________________| 
            center                   edge

            (o) Brightness  (o) Color [___]
            Curve File: [ Open ] [ Save ]

***/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);
   EFvignette.zd = zd;

   zdialog_add_widget(zd,"frame","frame","dialog",0,"expand");
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labcenter","hb1",ZTX("center"),"space=4");
   zdialog_add_widget(zd,"label","space","hb1",0,"expand");
   zdialog_add_widget(zd,"label","labedge","hb1",ZTX("edge"),"space=5");

   zdialog_add_widget(zd,"hbox","hb2","dialog",0,"space=3");
   zdialog_add_widget(zd,"radio","RBbrite","hb2",Bbrightness,"space=5");
   zdialog_add_widget(zd,"radio","RBcolor","hb2",Bcolor,"space=5");
   zdialog_add_widget(zd,"colorbutt","color","hb2","0|0|255");

   zdialog_add_widget(zd,"hbox","hb3","dialog",0,"space=5");
   zdialog_add_widget(zd,"label","labcurve","hb3",Bcurvefile,"space=5");
   zdialog_add_widget(zd,"button","load","hb3",Bopen,"space=5");
   zdialog_add_widget(zd,"button","save","hb3",Bsave,"space=5");
   
   vignette_RGB[0] = vignette_RGB[1] = 0;                                  //  initial color = blue
   vignette_RGB[2] = 255;
   
   vign_cx = E3pxm->ww / 2;                                                //  initial vignette center         v.12.10
   vign_cy = E3pxm->hh / 2;
   
   vign_rad = vign_cx * vign_cx + vign_cy * vign_cy;                       //  radius = distance to corners    v.12.10
   vign_rad = sqrtf(vign_rad);
   
   zdialog_stuff(zd,"RBbrite",1);                                          //  default curve = brightness

   GtkWidget *frame = zdialog_widget(zd,"frame");                          //  set up curve edit
   spldat *sd = splcurve_init(frame,Vign_curvedit);
   EFvignette.curves = sd;

   sd->vert[0] = 0;                                                        //  curve 0 = active curve = brightness
   sd->nap[0] = 2;
   sd->apx[0][0] = 0.01;
   sd->apy[0][0] = 0.5;
   sd->apx[0][1] = 0.99;
   sd->apy[0][1] = 0.5;
   splcurve_generate(sd,0);                                                //  generate curve data

   sd->vert[1] = 0;                                                        //  curve 1 = brightness curve
   sd->nap[1] = 2;
   sd->apx[1][0] = 0.01;
   sd->apy[1][0] = 0.5;
   sd->apx[1][1] = 0.99;
   sd->apy[1][1] = 0.5;
   splcurve_generate(sd,1);

   sd->vert[2] = 0;                                                        //  curve 2 = color curve
   sd->nap[2] = 2;
   sd->apx[2][0] = 0.01;
   sd->apy[2][0] = 0.01;
   sd->apx[2][1] = 0.99;
   sd->apy[2][1] = 0.01;
   splcurve_generate(sd,2);

   sd->Nspc = 1;                                                           //  only one active curve
   vignette_spc = 1;                                                       //  initial curve = brightness

   zdialog_resize(zd,300,300);
   zdialog_run(zd,Vign_dialog_event,"save");                               //  run dialog - parallel
   
   takeMouse(vign_mousefunc,dragcursor);                                   //  connect mouse function

   return;
}


//  dialog event and completion callback function

int Vign_dialog_event(zdialog *zd, cchar *event)
{
   void     Vign_curvedit(int);

   spldat   *sd = EFvignette.curves;
   int      ii, jj;
   char     color[20];
   cchar    *pp;

   if (zd->zstat)                                                          //  dialog complete
   {
      if (zd->zstat == 1) {                                                //  done
         float R = 1.0 * E0pxm->ww / E3pxm->ww;
         vign_cx = R * vign_cx;                                            //  scale geometries to full size   v.12.10
         vign_cy = R * vign_cy;
         vign_rad = R * vign_rad;
         edit_done();
      }
      else edit_cancel();                                                  //  cancel
      return 1;
   }

   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(vign_mousefunc,dragcursor);                                //  connect mouse function

   if (strnEqu(event,"RB",2)) {                                            //  new choice of curve
      ii = strcmpv(event,"RBbrite","RBcolor",null);
      vignette_spc = ii;
      sd->nap[0] = sd->nap[ii];                                            //  copy new curve to curve 0
      for (jj = 0; jj < sd->nap[0]; jj++) {
         sd->apx[0][jj] = sd->apx[ii][jj];
         sd->apy[0][jj] = sd->apy[ii][jj];
      }
      splcurve_generate(sd,0);                                             //  regenerate curve 0
      gtk_widget_queue_draw(sd->drawarea);                                 //  draw curve
      signal_thread();                                                     //  v.13.04
   }

   if (strEqu(event,"blendwidth")) signal_thread();

   if (strEqu(event,"color")) {                                            //  change color
      zdialog_fetch(zd,"color",color,19);                                  //  get color from color wheel
      pp = strField(color,"|",1);
      if (pp) vignette_RGB[0] = atoi(pp);
      pp = strField(color,"|",2);
      if (pp) vignette_RGB[1] = atoi(pp);
      pp = strField(color,"|",3);
      if (pp) vignette_RGB[2] = atoi(pp);
      signal_thread();                                                     //  trigger update thread
   }

   if (strEqu(event,"load")) {                                             //  load saved curve 
      splcurve_load(sd);
      Vign_curvedit(0);                                                    //  v.12.09
      signal_thread();
      return 0;
   }

   if (strEqu(event,"save")) {                                             //  save curve to file 
      splcurve_save(sd);
      return 0;
   }
   
   return 0;
}


//  get mouse position and set new center for vignette

void vign_mousefunc()                                                      //  mouse function               v.12.09
{
   if (! LMclick && ! Mdrag) return;
   LMclick = 0;
   
   vign_cx = Mxposn;                                                       //  new vignette center = mouse position
   vign_cy = Myposn;

   Mxdrag = Mydrag = 0;                                                    //  v.12.11

   signal_thread();                                                        //  trigger image update
   return;
}


//  this function is called when the curve is edited

void Vign_curvedit(int)
{
   int      ii = vignette_spc, jj;
   spldat   *sd = EFvignette.curves;

   sd->nap[ii] = sd->nap[0];                                               //  copy curve 0 to current curve
   for (jj = 0; jj < sd->nap[0]; jj++) {
      sd->apx[ii][jj] = sd->apx[0][jj];
      sd->apy[ii][jj] = sd->apy[0][jj];
   }
   for (jj = 0; jj < 1000; jj++)
      sd->yval[ii][jj] = sd->yval[0][jj];

   signal_thread();                                                        //  update image
   return;
}


//  thread function

void * Vign_thread(void *)
{
   void * Vign_wthread(void *arg);

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request

      for (int ii = 0; ii < Nwt; ii++)                                     //  start working threads
         start_wthread(Vign_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion
      
      CEF->Fmods++; 
      mwpaint2(); 
   }

   return 0;                                                               //  not executed, stop g++ warning
}


//  working thread

void * Vign_wthread(void *arg)
{
   float       *pix1, *pix3;
   int         index, ii, kk, px, py, dist = 0;
   float       cx, cy, rad, radx, rady, f1, f2, xval, yval;
   float       red1, green1, blue1, red3, green3, blue3, cmax;
   spldat      *sd = EFvignette.curves;

   cx = vign_cx;                                                           //  vignette center (mouse)
   cy = vign_cy;

   index = *((int *) arg);

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      ii = py * E3pxm->ww + px;

      if (sa_stat == 3) {                                                  //  select area active 
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  pixel is outside area
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel

      red1 = pix1[0];                                                      //  input RGB
      green1 = pix1[1];
      blue1 = pix1[2];

      radx = px - cx;                                                      //  distance from vignette center
      rady = py - cy;
      rad = sqrtf(radx*radx + rady*rady);                                  //  (px,py) distance from center
     
      xval = rad / vign_rad;                                               //  scale 0 to 1.0                  v.12.09
      kk = 999.0 * xval;                                                   //  scale 0 to 999
      if (kk > 999) kk = 999;                                              //  beyond radius                   v.12.10

      yval = sd->yval[1][kk];                                              //  brightness curve y-value 0 to 1.0
      if (yval > 1.0) yval = 1.0;
      yval = 2.0 * yval;                                                   //  0 to 2.0
      
      red3 = yval * red1;                                                  //  adjust brightness
      green3 = yval * green1;
      blue3 = yval * blue1;

      yval = sd->yval[2][kk];                                              //  color curve y-value 0 to 1.0
      if (yval > 1.0) yval = 1.0;
      f1 = yval;                                                           //  0 to 1.0   new color
      f2 = 1.0 - f1;                                                       //  1.0 to 0   old color

      red3 = f1 * vignette_RGB[0] + f2 * red3;                             //  mix input and vignette color
      green3 = f1 * vignette_RGB[1] + f2 * green3;
      blue3 = f1 * vignette_RGB[2] + f2 * blue3;

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         red3 = f1 * red3 + f2 * red1;
         green3 = f1 * green3 + f2 * green1;
         blue3 = f1 * blue3 + f2 * blue1;
      }
      
      cmax = red3;                                                         //  detect overflow
      if (green3 > cmax) cmax = green3;
      if (blue3 > cmax) cmax = blue3;
      
      if (cmax > 255.9) {                                                  //  stop overflow
         red3 = red3 * 255.9 / cmax;
         green3 = green3 * 255.9 / cmax;
         blue3 = blue3 * 255.9 / cmax;
      }

      pix3[0] = red3;
      pix3[1] = green3;
      pix3[2] = blue3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, stop g++ warning
}


/**************************************************************************/

//  match_color edit function
//  Adjust colors of image 2 to match the colors of image 1
//  using small selected areas in each image as the match standard.

void * match_color_thread(void *);
void   match_color_mousefunc();

float    match_color_RGB1[3];                                              //  image 1 base colors to match
float    match_color_RGB2[3];                                              //  image 2 target colors to match
int      match_color_radius = 10;                                          //  mouse radius
int      match_color_mode = 0;

editfunc    EFmatchcolor;


void m_match_color(GtkWidget *, const char *) 
{
   int    match_color_dialog_event(zdialog* zd, const char *event);
   
   cchar    *title = ZTX("Color Match Images");

   F1_help_topic = "match_colors";
   
   if (checkmenulock()) return;                                            //  test menu lock            v.12.09
   
/*
               Color Match Images
         1  [ 10 ]   mouse radius for color sample
         2  [Open]   image for source color
         3  click on image to get source color
         4  [Open]   image for target color
         5  click on image to set target color
                             [done] [cancel]
*/

   zdialog *zd = zdialog_new(title,mWin,Bdone,Bcancel,null);               //  match_color dialog
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=2");
   zdialog_add_widget(zd,"vbox","vb1","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"vbox","vb2","hb1",0,"homog|space=3");
   zdialog_add_widget(zd,"label","labn1","vb1","1");
   zdialog_add_widget(zd,"label","labn2","vb1","2");
   zdialog_add_widget(zd,"label","labn3","vb1","3");
   zdialog_add_widget(zd,"label","labn4","vb1","4");
   zdialog_add_widget(zd,"label","labn5","vb1","5");
   zdialog_add_widget(zd,"hbox","hbrad","vb2");
   zdialog_add_widget(zd,"spin","radius","hbrad","1|20|1|10","space=5");
   zdialog_add_widget(zd,"label","labrad","hbrad",ZTX("mouse radius for color sample"));
   zdialog_add_widget(zd,"hbox","hbop1","vb2");
   zdialog_add_widget(zd,"button","open1","hbop1",ZTX("Open"),"space=5");
   zdialog_add_widget(zd,"label","labop1","hbop1",ZTX("image for source color"));
   zdialog_add_widget(zd,"hbox","hbclik1","vb2");
   zdialog_add_widget(zd,"label","labclik1","hbclik1",ZTX("click on image to get source color"));
   zdialog_add_widget(zd,"hbox","hbop2","vb2");
   zdialog_add_widget(zd,"button","open2","hbop2",ZTX("Open"),"space=5");
   zdialog_add_widget(zd,"label","labop2","hbop2",ZTX("image to set matching color"));
   zdialog_add_widget(zd,"hbox","hbclik2","vb2");
   zdialog_add_widget(zd,"label","labclik2","hbclik2",ZTX("click on image to set matching color"));

   zdialog_stuff(zd,"radius",match_color_radius);                          //  remember last radius
   zdialog_run(zd,match_color_dialog_event);                               //  run dialog - parallel

   EFmatchcolor.funcname = "match-color";
   EFmatchcolor.Farea = 1;                                                 //  select area ignored
   EFmatchcolor.threadfunc = match_color_thread;
   EFmatchcolor.mousefunc = match_color_mousefunc;
   
   match_color_mode = 0;
   if (curr_file) {
      match_color_mode = 1;                                                //  image 1 ready to click
      takeMouse(match_color_mousefunc,0);                                  //  connect mouse function
   }

   return;
}


//  match_color dialog event and completion function

int match_color_dialog_event(zdialog *zd, const char *event)               //  match_color dialog event function
{
   int      err;
   
   if (zd->zstat)                                                          //  end dialog
   {
      if (match_color_mode == 4) {                                         //  edit was started
         if (zd->zstat == 1) edit_done();
         else edit_cancel();
      }
      freeMouse();
      zdialog_free(zd);
      match_color_mode = 0;
      return 1;
   }
   
   if (strEqu(event,"radius"))                                             //  set new mouse radius
      zdialog_fetch(zd,"radius",match_color_radius);
   
   if (strEqu(event,"open1"))                                              //  get image 1 for color source
   {
      match_color_mode = 0;
      err = f_open(null,0,0);
      if (! err) match_color_mode = 1;                                     //  image 1 ready to click
   }

   if (strEqu(event,"open2"))                                              //  get image 2 to set matching color
   {
      if (match_color_mode < 2) {
         zmessageACK(mWin,0,ZTX("select source image color first"));       //  check that RGB1 has been set
         return 1;
      }
      match_color_mode = 2;
      err = f_open(null,0,0);
      if (err) return 1;
      match_color_mode = 3;                                                //  image 2 ready to click
   }

   takeMouse(match_color_mousefunc,0);                                     //  reconnect mouse function
   return 1;
}


//  mouse function - click on image and get colors to match

void match_color_mousefunc()
{
   void  match_color_getRGB(int px, int py, float rgb[3]);

   int      px, py, mrect;

   if (match_color_mode < 1) return;                                       //  no image available yet
   
   mrect = 2 * match_color_radius;
   draw_mousearc(Mxposn,Myposn,mrect,mrect);
   
   if (LMclick)
   {
      LMclick = 0;
      px = Mxclick;
      py = Myclick;
   
      if (match_color_mode == 1 || match_color_mode == 2)                  //  image 1 ready to click
      {
         match_color_getRGB(px,py,match_color_RGB1);                       //  get RGB1 color
         match_color_mode = 2;
         return;
      }
      
      if (match_color_mode == 3 || match_color_mode == 4)                  //  image 2 ready to click
      {
         match_color_getRGB(px,py,match_color_RGB2);                       //  get RGB2 color

         if (match_color_mode == 4) {
            edit_cancel();                                                 //  if prior edit, cancel it
            takeMouse(match_color_mousefunc,0);                            //  reconnect mouse function
            match_color_mode = 3;
         }

         if (! edit_setup(EFmatchcolor)) return;                           //  setup edit - thread will launch
         match_color_mode = 4;                                             //  edit waiting for cancel or done
         signal_thread();                                                  //  update the target image
         return;
      }
   }

   return;
}


//  get the RGB averages for pixels within mouse radius

void match_color_getRGB(int px, int py, float rgb[3])
{
   int      rad1 = match_color_radius;
   int      rad2 = rad1 * rad1;
   int      rad, npix, qx, qy;
   float    red, green, blue;
   float    *pix1;
   PXM      *pxm;

   pxm = PXM_load(curr_file,1);                                            //  popup ACK if error     v.12.03
   if (! pxm) return;

   npix = 0;   
   red = green = blue = 0;

   for (qy = py-rad1; qy <= py+rad1; qy++)
   for (qx = px-rad1; qx <= px+rad1; qx++)
   {
      if (qx < 0 || qx > pxm->ww-1) continue;
      if (qy < 0 || qy > pxm->hh-1) continue;
      rad = (qx-px) * (qx-px) + (qy-py) * (qy-py);
      if (rad > rad2) continue;
      pix1 = PXMpix(pxm,qx,qy);
      red += pix1[0];
      green += pix1[1];
      blue += pix1[2];
      npix++;
   }
   
   rgb[0] = red / npix;
   rgb[1] = green / npix;
   rgb[2] = blue / npix;
   
   PXM_free(pxm);
   return;
}


//  thread function - start multiple working threads

void * match_color_thread(void *)
{
   void  * match_color_wthread(void *arg);                                 //  worker thread
   
   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(match_color_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++; 
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, avoid warning
}


void * match_color_wthread(void *arg)                                      //  worker thread function
{
   int         index = *((int *) (arg));
   int         px, py;
   float       *pix3;
   float       Rred, Rgreen, Rblue;
   float       red, green, blue, cmax;
   
   Rred = match_color_RGB1[0] / match_color_RGB2[0];                       //  color adjustment ratios
   Rgreen = match_color_RGB1[1] / match_color_RGB2[1];
   Rblue = match_color_RGB1[2] / match_color_RGB2[2];

   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      pix3 = PXMpix(E3pxm,px,py);
      
      red = pix3[0] * Rred;                                                //  adjust colors
      green = pix3[1] * Rgreen;
      blue = pix3[2] * Rblue;
      
      cmax = red;                                                          //  check for overflow
      if (green > cmax) cmax = green;
      if (blue > cmax) cmax = blue;

      if (cmax > 255.9) {                                                  //  fix overflow
         red = red * 255.9 / cmax;
         green = green * 255.9 / cmax;
         blue = blue * 255.9 / cmax;
      }      

      pix3[0] = red;
      pix3[1] = green;
      pix3[2] = blue;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  Show RGB values for 1-9 pixels selected with mouse-clicks.
//  Revise RGB values for the selected points and then revise all other 
//  pixels to match, using a distance-weighted average of the selected pixels.

void  RGBR_dialog();
void  RGBR_mousefunc();
void  RGBR_stuff();
void * RGBR_thread(void *);

zdialog  *RGBRzd; 
int      RGBRpixel[9][2];                                                  //  last 9 pixel locations clicked
float    RGBRval1[9][3];                                                   //  original RGB values, 0 to 255.9
float    RGBRval3[9][3];                                                   //  revised RGB values, 0 to 255.9
int      RGBRmetric = 1;                                                   //  1/2/3 = /RGB/EV/OD
int      RGBRdelta = 0;                                                    //  flag, delta mode
int      RGBRrestart;                                                      //  flag, restart dialog
int      RGBRnpix;                                                         //  no. of selected pixels
double   RGBRtime;                                                         //  last click time
int      RGBRblend;                                                        //  blend factor, 1 to 100
editfunc EFrgbrevise;                                                      //  edit function parameters

#define RGB2EV(rgb) (log2(rgb)-7)                                          //  RGB 0.1 to 255.9 >> EV -10 to +1
#define EV2RGB(ev) (pow(2,ev+7))                                           //  inverse
#define RGB2OD(rgb) (2-log10(100*rgb/256))                                 //  RGB 0.1 to 255.9 >> OD 3.4 to 0.000017
#define OD2RGB(od) (pow(10,2.40824-od))                                    //  inverse


void m_revise_RGB(GtkWidget *, cchar *menu)
{
   int   RGBR_event(zdialog *zd, cchar *event);

   GtkWidget   *widget;
   cchar       *limits;
   zdialog     *zd;
   cchar       *mess = ZTX("Click image to select pixels.");

   PangoFontDescription    *widgetfont;
   widgetfont = pango_font_description_from_string("Monospace 8");         //  small monospace font for widgets

   F1_help_topic = "revise_RGB";
   
   EFrgbrevise.funcname = "revise_RGB";                                    //  setup edit function
   EFrgbrevise.threadfunc = RGBR_thread;
   EFrgbrevise.mousefunc = RGBR_mousefunc;
   EFrgbrevise.Farea = 1;
   if (! edit_setup(EFrgbrevise)) return;

   RGBRblend = 20;                                                         //  default slider value
   RGBRnpix = 0;                                                           //  no pixels selected yet
   
/***
   
      Click image to select pixels.
      [x] delta
      Metric: (o) RGB  (o) EV  (o) OD
      Pixel          Red      Green    Blue
      A xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v]                             //  spin buttons for RGB/EV/OD values
      B xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v]
      C xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      D xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      E xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      F xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      G xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      H xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      I xxxx xxxx   [xxxx|v] [xxxx|v] [xxxx|v] 
      Blend  ==============[]================

                     [reset] [done] [cancel]
***/
   
restart:

   zd = zdialog_new(ZTX("Revise RGB"),mWin,Breset,Bdone,Bcancel,null);
   EFrgbrevise.zd = zd;
   RGBRzd = zd;

   zdialog_add_widget(zd,"hbox","hbmess","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmess","hbmess",mess,"space=5");

   zdialog_add_widget(zd,"hbox","hbmym","dialog");
   zdialog_add_widget(zd,"check","delta","hbmym","delta","space=5");
   zdialog_stuff(zd,"delta",RGBRdelta);

   zdialog_add_widget(zd,"hbox","hbmetr","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","labmetr","hbmetr",ZTX("Metric:"),"space=5");
   zdialog_add_widget(zd,"radio","radRGB","hbmetr","RGB","space=3");
   zdialog_add_widget(zd,"radio","radEV","hbmetr","EV","space=3");
   zdialog_add_widget(zd,"radio","radOD","hbmetr","OD","space=3");
   
   if (RGBRmetric == 1) zdialog_stuff(zd,"radRGB",1);                      //  set current metric
   if (RGBRmetric == 2) zdialog_stuff(zd,"radEV",1);
   if (RGBRmetric == 3) zdialog_stuff(zd,"radOD",1);

   zdialog_add_widget(zd,"hbox","hbdata","dialog");
   zdialog_add_widget(zd,"vbox","vbpix","hbdata",0,"space=3|homog");       //  vbox for pixel locations
   zdialog_add_widget(zd,"hbox","hbpix","vbpix");
   zdialog_add_widget(zd,"label","labpix","hbpix","Pixel");                //  header
   
   char hbpixx[8] = "hbpixx", pixx[8] = "pixx";

   for (int ii = 1; ii < 10; ii++) {                                       //  add labels for pixel locations
      hbpixx[5] = '0' + ii;
      pixx[3] = '0' + ii;
      zdialog_add_widget(zd,"hbox",hbpixx,"vbpix");                        //  add hbox "hbpix1" to "hbpix9"
      zdialog_add_widget(zd,"label",pixx,hbpixx);                          //  add label "pix1" to "pix9"
      widget = zdialog_widget(zd,pixx);
      gtk_widget_modify_font(widget,widgetfont);                           //  use monofont
   }

   if (RGBRmetric == 1) limits = "-255.9|255.9|0.1|0.0";                   //  metric = RGB
   else if (RGBRmetric == 2) limits = "-8|8|0.01|0.0";                     //           EV
   else limits = "-3|3|0.002|0.0";                                         //           OD

   zdialog_add_widget(zd,"vbox","vbdat","hbdata",0,"space=3|homog");       //  vbox for pixel RGB values
   zdialog_add_widget(zd,"hbox","hbrgb","vbdat",0,"homog");
   zdialog_add_widget(zd,"label","labr","hbrgb",Bred);
   zdialog_add_widget(zd,"label","labg","hbrgb",Bgreen);
   zdialog_add_widget(zd,"label","labb","hbrgb",Bblue);

   char hbdatx[8] = "hbdatx", redx[8] = "redx", greenx[8] = "greenx", bluex[8] = "bluex";

   for (int ii = 1; ii < 10; ii++) {
      hbdatx[5] = '0' + ii;
      redx[3] = '0' + ii;
      greenx[5] = '0' + ii;
      bluex[4] = '0' + ii;
      zdialog_add_widget(zd,"hbox",hbdatx,"vbdat");                        //  add hbox "hbdat1" to "hbdat9"
      zdialog_add_widget(zd,"spin",redx,hbdatx,limits,"space=3");          //  add spin buttons for "red1" to "red9" etc.
      zdialog_add_widget(zd,"spin",greenx,hbdatx,limits,"space=3");
      zdialog_add_widget(zd,"spin",bluex,hbdatx,limits,"space=3");
      widget = zdialog_widget(zd,redx);
      gtk_widget_modify_font(widget,widgetfont);                           //  use monofont
      widget = zdialog_widget(zd,greenx);
      gtk_widget_modify_font(widget,widgetfont);
      widget = zdialog_widget(zd,bluex);
      gtk_widget_modify_font(widget,widgetfont);
   }
   
   zdialog_add_widget(zd,"hbox","hbsoft","dialog","space=5");
   zdialog_add_widget(zd,"label","labsoft","hbsoft",ZTX("Blend"),"space=3");
   zdialog_add_widget(zd,"hscale","blend","hbsoft","0|100|1|20","space=5|expand");
   zdialog_add_widget(zd,"label","softval","hbsoft","20");

   RGBR_stuff();                                                           //  stuff current pixels (if restart)

   takeMouse(RGBR_mousefunc,dragcursor);                                   //  connect mouse function
   zdialog_run(zd,RGBR_event,"save");                                      //  run dialog
   zdialog_wait(zd);                                                       //  wait for completion
   if (RGBRrestart) goto restart;                                          //  restart dialog

   return;
}


//  dialog event function

int RGBR_event(zdialog *zd, cchar *event)
{
   int         button;
   int         ii, jj;
   char        text[8];
   float       val1, val3;
   
   if (zd->zstat) 
   {
      if (zd->zstat == 1) {                                                //  Reset
         zd->zstat = 0;                                                    //  keep dialog active
         RGBRnpix = 0;                                                     //  no pixels selected yet
         RGBR_stuff();                                                     //  clear dialog
         edit_reset();                                                     //  reset edits 
         return 0;
      }
      else if (zd->zstat == 2) edit_done();
      else edit_cancel();                                                  //  Cancel or [x]
      freeMouse();                                                         //  disconnect mouse function
      zdialog_free(zd);                                                    //  kill dialog
      RGBRzd = 0;
      RGBRrestart = 0;
      erase_toptext(101);                                                  //  erase pixel labels from image
      return 1;
   }
   
   if (strEqu(event,"focus"))                                              //  toggle mouse capture
      takeMouse(RGBR_mousefunc,dragcursor);                                //  connect mouse function
   
   if (strEqu(event,"blend")) {                                            //  new blend factor
      zdialog_fetch(zd,"blend",RGBRblend);
      sprintf(text,"%d",RGBRblend);                                        //  numeric feedback
      zdialog_stuff(zd,"softval",text);
      signal_thread();
      return 1;
   }

   if (strnEqu(event,"rad",3))                                             //  metric was changed
   {
      if (strEqu(event,"radRGB")) {                                        //  RGB
         zdialog_fetch(zd,event,button);
         if (button) RGBRmetric = 1;
      }
         
      if (strEqu(event,"radEV")) {                                         //  EV
         zdialog_fetch(zd,event,button);
         if (button) RGBRmetric = 2;
      }
         
      if (strEqu(event,"radOD")) {                                         //  OD
         zdialog_fetch(zd,event,button);
         if (button) RGBRmetric = 3;
      }

      if (button) {      
         freeMouse();                                                      //  restart dialog with new limits
         zdialog_free(zd);
         RGBRzd = 0;
         RGBRrestart = 1;
      }

      return 1;
   }

   if (strEqu(event,"delta")) {                                            //  change absolute/delta mode
      zdialog_fetch(zd,"delta",RGBRdelta);
      freeMouse();                                                         //  restart dialog with new limits
      zdialog_free(zd);
      RGBRzd = 0;
      RGBRrestart = 1;
      return 1;
   }
   
   ii = -1;                                                                //  no RGB change yet

   if (strnEqu(event,"red",3)) {                                           //  red1 - red9 was changed
      ii = event[3] - '1';                                                 //  pixel index 0-8
      jj = 0;                                                              //  color = red
   }

   if (strnEqu(event,"green",5)) {                                         //  green1 - green9
      ii = event[5] - '1';
      jj = 1;
   }

   if (strnEqu(event,"blue",4)) {                                          //  blue1 - blue9
      ii = event[4] - '1';
      jj = 2;
   }
   
   if (ii >= 0 && ii < 9)                                                  //  RGB value was revised
   {
      val1 = RGBRval1[ii][jj];                                             //  original pixel RGB value
      if (RGBRmetric == 2) val1 = RGB2EV(val1);                            //  convert to EV or OD units
      if (RGBRmetric == 3) val1 = RGB2OD(val1);

      zdialog_fetch(zd,event,val3);                                        //  revised RGB/EV/OD value from dialog
      if (RGBRdelta) val3 += val1;                                         //  if delta mode, make absolute

      if (RGBRmetric == 2) val3 = EV2RGB(val3);                            //  convert EV/OD to RGB value
      if (RGBRmetric == 3) val3 = OD2RGB(val3);

      if (fabsf(RGBRval3[ii][jj] - val3) < 0.001) return 1;                //  ignore re-entry after change

      if (val3 < 0.1) val3 = 0.1;                                          //  limit RGB within 0.1 to 255.9
      if (val3 > 255.9) val3 = 255.9;

      RGBRval3[ii][jj] = val3;                                             //  new RGB value for pixel
      
      if (RGBRmetric == 2) val3 = RGB2EV(val3);                            //  convert RGB to EV/OD units
      if (RGBRmetric == 3) val3 = RGB2OD(val3);

      if (RGBRdelta) val3 -= val1;                                         //  absolute back to relative
      zdialog_stuff(zd,event,val3);                                        //  limited value back to dialog
      
      signal_thread();                                                     //  signal thread to update image
   }

   return 1;
}


//  mouse function

void RGBR_mousefunc()                                                      //  mouse function
{
   int         ii;
   float       *pix3;

   if (! LMclick) return;
   LMclick = 0;

   RGBRtime = get_seconds();                                               //  mark time of pixel click
   
   if (RGBRnpix == 9) {                                                    //  if table is full (9 entries)
      for (ii = 1; ii < 9; ii++) {                                         //    move positions 1-8 up 
         RGBRpixel[ii-1][0] = RGBRpixel[ii][0];                            //      to fill positions 0-7
         RGBRpixel[ii-1][1] = RGBRpixel[ii][1];
      }
      RGBRnpix = 8;                                                        //  count is now 8 entries
   }
   
   ii = RGBRnpix;                                                          //  next table position to fill
   RGBRpixel[ii][0] = Mxclick;                                             //  newest pixel
   RGBRpixel[ii][1] = Myclick;
   
   pix3 = PXMpix(E3pxm,Mxclick,Myclick);                                   //  get initial RGB values from
   RGBRval1[ii][0] = RGBRval3[ii][0] = pix3[0];                            //    modified image E3
   RGBRval1[ii][1] = RGBRval3[ii][1] = pix3[1];
   RGBRval1[ii][2] = RGBRval3[ii][2] = pix3[2];

   RGBRnpix++;                                                             //  up pixel count
   RGBR_stuff();                                                           //  stuff pixels and values into dialog
   return;
}


//  stuff dialog with current pixels and their RGB/EV/OD values

void RGBR_stuff()
{
   static char    lab1[9][4] = { "A", "B", "C", "D", "E", "F", "G", "H", "I" };
   static char    lab2[9][4] = { " A ", " B ", " C ", " D ", " E ", " F ", " G ", " H ", " I " };

   int         px, py;
   float       red1, green1, blue1, red3, green3, blue3;
   char        text[100], pixx[8] = "pixx";
   char        redx[8] = "redx", greenx[8] = "greenx", bluex[8] = "bluex";
   zdialog     *zd = RGBRzd;
   
   erase_toptext(101);                                                     //  erase prior labels from image
   
   for (int ii = 0; ii < 9; ii++)                                          //  loop slots 0-8
   {
      pixx[3] = '1' + ii;                                                  //  widget names "pix1" ... "pix9"
      redx[3] = '1' + ii;                                                  //  widget names "red1" ... "red9"
      greenx[5] = '1' + ii;
      bluex[4] = '1' + ii;

      px = RGBRpixel[ii][0];                                               //  next pixel to report
      py = RGBRpixel[ii][1];
      if (ii >= RGBRnpix) {                                                //  > last pixel selected
         zdialog_stuff(zd,pixx,"");
         zdialog_stuff(zd,redx,0);                                         //  blank pixel and zero values
         zdialog_stuff(zd,greenx,0);
         zdialog_stuff(zd,bluex,0);
         continue;
      }

      sprintf(text,"%s %4d %4d",lab1[ii],px,py);                           //  format pixel "A nnnn nnnn"
      zdialog_stuff(zd,pixx,text);                                         //  pixel >> widget
      
      add_toptext(101,px,py,lab2[ii],"Sans 8");                            //  paint label on image at pixel
      
      red1 = RGBRval1[ii][0];                                              //  original RGB values for pixel
      green1 = RGBRval1[ii][1];
      blue1 = RGBRval1[ii][2];
      red3 = RGBRval3[ii][0];                                              //  revised RGB values
      green3 = RGBRval3[ii][1];
      blue3 = RGBRval3[ii][2];
      
      if (RGBRmetric == 2) {                                               //  convert to EV units if needed
         red1 = RGB2EV(red1);
         green1 = RGB2EV(green1);
         blue1 = RGB2EV(blue1);
         red3 = RGB2EV(red3);
         green3 = RGB2EV(green3);
         blue3 = RGB2EV(blue3);
      }
      
      if (RGBRmetric == 3) {                                               //  or OD units
         red1 = RGB2OD(red1);
         green1 = RGB2OD(green1);
         blue1 = RGB2OD(blue1);
         red3 = RGB2OD(red3);
         green3 = RGB2OD(green3);
         blue3 = RGB2OD(blue3);
      }
      
      if (RGBRdelta) {                                                     //  dialog is delta mode
         red3 -= red1;
         green3 -= green1;
         blue3 -= blue1;
      }
      
      zdialog_stuff(zd,redx,red3);
      zdialog_stuff(zd,greenx,green3);
      zdialog_stuff(zd,bluex,blue3);
   }

   mwpaint2();                                                             //  refresh window
   return;
}


//  thread function - multiple working threads to update image

void * RGBR_thread(void *)
{
   void  * RGBR_wthread(void *arg);                                        //  worker thread

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      zsleep(0.3);                                                         //  more time for dialog controls
      
      if (RGBRnpix < 1) continue;                                          //  must have 1+ pixels in table
      
      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(RGBR_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      CEF->Fmods++;                                                        //  image modified
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop warning
}


void * RGBR_wthread(void *arg)                                             //  worker thread function
{
   int         index = *((int *) (arg));
   int         px1, py1, px2, py2, ii;
   float       *pix1, *pix3;
   float       dist[9], weight[9], sumdist;
   float       blend, delta, red, green, blue, max;
   
   blend = E1pxm->ww;
   if (E1pxm->hh > E1pxm->ww) blend = E1pxm->hh;
   blend = blend * blend;
   blend = 0.0002 * RGBRblend * RGBRblend * blend + 100;                   //  100 to 2 * E1pxm->ww**2
   
   for (py1 = index; py1 < E3pxm->hh; py1 += Nwt)                          //  loop all image pixels
   for (px1 = 0; px1 < E3pxm->ww; px1++)
   {
      for (sumdist = ii = 0; ii < RGBRnpix; ii++)                          //  compute weight of each revision point
      {
         px2 = RGBRpixel[ii][0];
         py2 = RGBRpixel[ii][1];
         dist[ii] = (px1-px2)*(px1-px2) + (py1-py2)*(py1-py2);             //  distance (px1,py1) to (px2,py2)
         dist[ii] = 1.0 / (dist[ii] + blend);                              //  blend reduces peaks at revision points
         sumdist += dist[ii];                                              //  sum inverse distances
      }
      
      for (ii = 0; ii < RGBRnpix; ii++)                                    //  weight of each point
         weight[ii] = dist[ii] / sumdist;

      pix1 = PXMpix(E1pxm,px1,py1);                                        //  input pixel
      pix3 = PXMpix(E3pxm,px1,py1);                                        //  output pixel
      
      red = pix1[0];
      green = pix1[1];
      blue = pix1[2];
      
      for (ii = 0; ii < RGBRnpix; ii++) {                                  //  apply weighted color changes
         delta = RGBRval3[ii][0] - RGBRval1[ii][0];                        //    to each color
         red += weight[ii] * delta;
         delta = RGBRval3[ii][1] - RGBRval1[ii][1];
         green += weight[ii] * delta;
         delta = RGBRval3[ii][2] - RGBRval1[ii][2];
         blue += weight[ii] * delta;
      }
      
      max = red;
      if (green > max) max = green;
      if (blue > max) max = blue;

      if (max > 255.9) {                                                   //  stop overflow/underflow
         red = red * 255.9 / max;
         green = green * 255.9 / max;
         blue = blue * 255.9 / max;
      }
      
      if (red < 0) red = 0;
      if (green < 0) green = 0;
      if (blue < 0) blue = 0;
      
      pix3[0] = red;
      pix3[1] = green;
      pix3[2] = blue;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}


/**************************************************************************/

//  CMYK menu function
//  Adjust Density (Brightness) and RGB/CMY levels using 0.01 OD units

editfunc    EFcmyk;                                                        //  edit function data
int         CMYKinputs[8];
int         CMYKbias = 0;
char        *CMYK_file = 0;

void m_CMYK(GtkWidget *, const char *)
{
   int    CMYK_dialog_event(zdialog *zd, cchar *event);
   void * CMYK_thread(void *);

   F1_help_topic = "CMYK";
   
   EFcmyk.funcname = "CMYK";                                               //  function name
   EFcmyk.Fprev = 1;                                                       //  use preview
   EFcmyk.Farea = 2;                                                       //  select area usable
   EFcmyk.menufunc = m_CMYK;                                               //  allow restart                v.13.04
   EFcmyk.threadfunc = CMYK_thread;                                        //  thread function
   if (! edit_setup(EFcmyk)) return;                                       //  setup edit
   
/***
    _________________________________________________
   |                                                 |
   |  [reset]   [x] Add standard bias                |
   |  +Brightness -Density [__|v]   Contrast [__|v]  |
   |  +Red -Cyan           [__|v]     Red    [__|v]  |
   |  +Green -Magenta      [__|v]     Green  [__|v]  |
   |  +Blue -Yellow        [__|v]     Blue   [__|x]  |
   |                                                 |
   |                  [open] [save] [done] [cancel]  |
   |_________________________________________________|
   
***/

   zdialog *zd = zdialog_new("CMYK",mWin,Bopen,Bsave,Bdone,Bcancel,null);  //  new dialog
   EFcmyk.zd = zd;                                                         //  add dialog to edit function

   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=5");
   zdialog_add_widget(zd,"button","reset","hb1",Breset,"space=10");
   zdialog_add_widget(zd,"check","bias","hb1",ZTX("Add standard bias"),"space=5");
   zdialog_add_widget(zd,"hbox","hb2","dialog");
   zdialog_add_widget(zd,"vbox","vb1","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb2","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"label","space","hb2",0,"space=8");
   zdialog_add_widget(zd,"vbox","vb3","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"vbox","vb4","hb2",0,"homog|space=5");
   zdialog_add_widget(zd,"button","buBriteDens","vb1",ZTX("+Brightness -Density"));
   zdialog_add_widget(zd,"button","buRedDens","vb1",ZTX("+Red -Cyan"));
   zdialog_add_widget(zd,"button","buGreenDens","vb1",ZTX("+Green -Magenta"));
   zdialog_add_widget(zd,"button","buBlueDens","vb1",ZTX("+Blue -Yellow"));
   zdialog_add_widget(zd,"spin","spBriteDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spRedDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spGreenDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spBlueDens","vb2","-99|+99|1|0");
   zdialog_add_widget(zd,"button","buContrast","vb3",ZTX("Contrast"));
   zdialog_add_widget(zd,"button","buRedCon","vb3",ZTX("Red"));
   zdialog_add_widget(zd,"button","buGreenCon","vb3",ZTX("Green"));
   zdialog_add_widget(zd,"button","buBlueCon","vb3",ZTX("Blue"));
   zdialog_add_widget(zd,"spin","spContrast","vb4","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spRedCon","vb4","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spGreenCon","vb4","-99|+99|1|0");
   zdialog_add_widget(zd,"spin","spBlueCon","vb4","-99|+99|1|0");

   if (! CMYK_file) {                                                      //  default file
      CMYK_file = zmalloc(200,"CMYK");                                     //  /.../.fotoxx/CMYK_parameters
      snprintf(CMYK_file,200,"%s/CMYK_parameters",get_zuserdir());
   }

   zdialog_run(zd,CMYK_dialog_event,"save");                               //  run dialog - parallel
   return;
}


//  CMYK dialog event and completion function

int CMYK_dialog_event(zdialog *zd, cchar *event)                           //  CMYK dialog event function
{
   void  CMYK_open();
   void  CMYK_save();
   
   int      bias, mod;
   
   if (zd->zstat)
   {
      if (zd->zstat == 1) {                                                //  open file
         zd->zstat = 0;                                                    //  keep dialog active
         CMYK_open();                                                      //  read file

         zdialog_stuff(zd,"bias",CMYKbias);                                //  set all data
         zdialog_stuff(zd,"spBriteDens",CMYKinputs[0]);
         zdialog_stuff(zd,"spRedDens",CMYKinputs[1]);
         zdialog_stuff(zd,"spGreenDens",CMYKinputs[2]);
         zdialog_stuff(zd,"spBlueDens",CMYKinputs[3]);
         zdialog_stuff(zd,"spContrast",CMYKinputs[4]);
         zdialog_stuff(zd,"spRedCon",CMYKinputs[5]);
         zdialog_stuff(zd,"spGreenCon",CMYKinputs[6]);
         zdialog_stuff(zd,"spBlueCon",CMYKinputs[7]);
      }

      else if (zd->zstat == 2) {                                           //  save file
         zd->zstat = 0;                                                    //  keep dialog active
         CMYK_save();                                                      //  write file
      }

      else if (zd->zstat == 3) edit_done();                                //  done
      else edit_cancel();                                                  //  cancel or destroy
      return 1;
   }
   
   if (strEqu(event,"save")) {                                             //  from f_save()                v.13.04
      edit_fullsize();
      return 1;
   }
   
   if (strstr("focus",event)) return 1;

   if (strEqu(event,"bias"))                                               //  bias was changed
   {
      zdialog_fetch(zd,"bias",bias);
      if (bias != CMYKbias) 
      {
         if (bias) {                                                       //  off to on
            CMYKinputs[0] += 72;
            CMYKinputs[2] += 32;
            CMYKinputs[3] += 32;
            zdialog_set_limits(zd,"spBriteDens",-27,+171);
            zdialog_set_limits(zd,"spGreenDens",-67,+131);
            zdialog_set_limits(zd,"spBlueDens",-67,+131);
         }
         else {                                                            //  on to off
            CMYKinputs[0] -= 72;
            CMYKinputs[2] -= 32;
            CMYKinputs[3] -= 32;
            zdialog_set_limits(zd,"spBriteDens",-99,+99);
            zdialog_set_limits(zd,"spGreenDens",-99,+99);
            zdialog_set_limits(zd,"spBlueDens",-99,+99);
         }
         zdialog_stuff(zd,"spBriteDens",CMYKinputs[0]);
         zdialog_stuff(zd,"spGreenDens",CMYKinputs[2]);
         zdialog_stuff(zd,"spBlueDens",CMYKinputs[3]);
      }
      CMYKbias = bias;
      return 1;
   }
   
   if (strEqu(event,"reset"))                                              //  reset dialog controls
   {
      CMYKinputs[0] = CMYKbias * 72;
      CMYKinputs[1] = CMYKbias * 0;
      CMYKinputs[2] = CMYKbias * 32;
      CMYKinputs[3] = CMYKbias * 32;
      CMYKinputs[4] = 0;
      CMYKinputs[5] = 0;
      CMYKinputs[6] = 0;
      CMYKinputs[7] = 0;
      zdialog_stuff(zd,"spBriteDens",CMYKinputs[0]);
      zdialog_stuff(zd,"spRedDens",CMYKinputs[1]);
      zdialog_stuff(zd,"spGreenDens",CMYKinputs[2]);
      zdialog_stuff(zd,"spBlueDens",CMYKinputs[3]);
      zdialog_stuff(zd,"spContrast",CMYKinputs[4]);
      zdialog_stuff(zd,"spRedCon",CMYKinputs[5]);
      zdialog_stuff(zd,"spGreenCon",CMYKinputs[6]);
      zdialog_stuff(zd,"spBlueCon",CMYKinputs[7]);
   }

   zdialog_fetch(zd,"spBriteDens",CMYKinputs[0]);                          //  get all inputs
   zdialog_fetch(zd,"spRedDens",CMYKinputs[1]);
   zdialog_fetch(zd,"spGreenDens",CMYKinputs[2]);
   zdialog_fetch(zd,"spBlueDens",CMYKinputs[3]);
   zdialog_fetch(zd,"spContrast",CMYKinputs[4]);
   zdialog_fetch(zd,"spRedCon",CMYKinputs[5]);
   zdialog_fetch(zd,"spGreenCon",CMYKinputs[6]);
   zdialog_fetch(zd,"spBlueCon",CMYKinputs[7]);
   
   mod = 0;
   if (CMYKinputs[0] - CMYKbias * 72) mod++;
   if (CMYKinputs[1] - CMYKbias * 0) mod++;
   if (CMYKinputs[2] - CMYKbias * 32) mod++;
   if (CMYKinputs[3] - CMYKbias * 32) mod++;
   if (CMYKinputs[4]) mod++;
   if (CMYKinputs[5]) mod++;
   if (CMYKinputs[6]) mod++;
   if (CMYKinputs[7]) mod++;
   
   if (mod) signal_thread();                                               //  trigger update thread

   CEF->Fmods = mod;
   return 1;
}


//  load CMYK parameters from a previously saved file

void CMYK_open()
{
   int  CMYK_open_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   
   zd = zdialog_new(ZTX("Load CMYK parameters"),mWin,Bopen,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("File:"),"space=3");
   zdialog_add_widget(zd,"entry","file","hb1",0,"expand|scc=40");
   zdialog_add_widget(zd,"button","browse","hb1",Bbrowse,"space=5");
   
   zdialog_stuff(zd,"file",CMYK_file);
   
   zdialog_run(zd,CMYK_open_dialog_event);
   zdialog_wait(zd);
   return;
}


int CMYK_open_dialog_event(zdialog *zd, cchar *event)
{
   char     *pp, file[200];
   FILE     *fid = 0;
   int      zstat, nn;
   
   if (strEqu(event,"browse")) {
      pp = zgetfile1(ZTX("CMYK parameters file"),"open",CMYK_file);
      if (! pp) return 0;
      zdialog_stuff(zd,"file",pp);
      zfree(pp);
   }
   
   zstat = zd->zstat;                                                      //  completion button
   
   if (zstat)
   {
      if (zstat == 1)                                                      //  open
      {
         zdialog_fetch(zd,"file",file,200);                                //  get file from dialog
         if (CMYK_file) zfree(CMYK_file);
         CMYK_file = strdupz(file,0,"CMYK");

         fid = fopen(file,"r");                                            //  open file
         if (! fid) {
            zmessageACK(mWin,0,ZTX("file not found"));
            return 1;
         }
         
         nn = fscanf(fid,"CMYK parameters");                               //  read header
         
         nn = fscanf(fid,"%4d %4d %4d %4d %4d %4d %4d %4d %4d",            //  read data
                           &CMYKbias, 
                           &CMYKinputs[0], &CMYKinputs[1], &CMYKinputs[2],
                           &CMYKinputs[3], &CMYKinputs[4], &CMYKinputs[5], 
                           &CMYKinputs[6], &CMYKinputs[7]);

         if (nn != 9) zmessageACK(mWin,0,"file format error");

         fclose(fid);
         zdialog_free(zd);
      }
      
      else zdialog_free(zd);                                               //  cancel or [x]
   }

   return 1;
}


//  save CMYK parameters to a file

void CMYK_save()
{
   int  CMYK_save_dialog_event(zdialog *zd, cchar *event);

   zdialog     *zd;
   
   zd = zdialog_new(ZTX("Save CMYK parameters"),mWin,Bsave,Bcancel,null);
   zdialog_add_widget(zd,"hbox","hb1","dialog",0,"space=3");
   zdialog_add_widget(zd,"label","lab1","hb1",ZTX("File:"),"space=3");
   zdialog_add_widget(zd,"entry","file","hb1",0,"expand|scc=40");
   zdialog_add_widget(zd,"button","browse","hb1",Bbrowse,"space=5");
   
   zdialog_stuff(zd,"file",CMYK_file);
   
   zdialog_run(zd,CMYK_save_dialog_event);
   zdialog_wait(zd);

   return;
}


int CMYK_save_dialog_event(zdialog *zd, cchar *event)
{
   char     *pp, file[200];
   FILE     *fid = 0;
   int      zstat;
   
   if (strEqu(event,"browse")) {
      pp = zgetfile1(ZTX("CMYK parameters file"),"save",CMYK_file);
      if (! pp) return 0;
      zdialog_stuff(zd,"file",pp);
      zfree(pp);
   }
   
   zstat = zd->zstat;                                                      //  completion button
   
   if (zstat)
   {
      if (zstat == 1)                                                      //  save
      {
         zdialog_fetch(zd,"file",file,200);                                //  get file from dialog
         if (CMYK_file) zfree(CMYK_file);
         CMYK_file = strdupz(file,0,"CMYK");

         fid = fopen(file,"w");                                            //  open file
         if (! fid) {
            zmessageACK(mWin,0,strerror(errno));
            return 1;
         }
         
         fprintf(fid,"CMYK parameters \n");                                //  write headers

         fprintf(fid,"%4d %4d %4d %4d %4d %4d %4d %4d %4d ",               //  write data
                              CMYKbias, 
                              CMYKinputs[0], CMYKinputs[1], CMYKinputs[2],
                              CMYKinputs[3], CMYKinputs[4], CMYKinputs[5], 
                              CMYKinputs[6], CMYKinputs[7]);
         fprintf(fid,"\n");
         fclose(fid);
         zdialog_free(zd);
      }

      else zdialog_free(zd);                                               //  cancel or [x]
   }

   return 1;
}


//  thread function - multiple working threads to update image

void * CMYK_thread(void *)
{
   void  * CMYK_wthread(void *arg);                                        //  worker thread

   while (true)
   {
      thread_idle_loop();                                                  //  wait for work or exit request
      
      if (sa_stat == 3) Fbusy_goal = sa_Npixel;                            //  set up progress monitor
      else  Fbusy_goal = E3pxm->ww * E3pxm->hh;
      Fbusy_done = 0;

      for (int ii = 0; ii < Nwt; ii++)                                     //  start worker threads
         start_wthread(CMYK_wthread,&wtnx[ii]);
      wait_wthreads();                                                     //  wait for completion

      Fbusy_goal = Fbusy_done = 0;
      CEF->Fmods++;                                                        //  image modified
      mwpaint2();                                                          //  update window
   }

   return 0;                                                               //  not executed, stop warning
}


void * CMYK_wthread(void *arg)                                             //  worker thread function
{
   float       Du, Ru, Gu, Bu, Rd, Gd, Bd;
   float       con, conR, conG, conB;
   float       R1, G1, B1, R3, G3, B3;

   int         index = *((int *) (arg));
   int         px, py, ii, dist = 0;
   float       *pix1, *pix3;
   float       cmax, F, f1, f2;
   float       c1 = 100.0 / 256.0;
   float       c2 = 2.0 + log10(1.0/c1);
   float       c3 = 1.0 / 256.0;
   
   #define ODfunc(rgb) (2.0 - log10(c1 * rgb))                             //  RGB units (1-256) to OD units
   #define RGBfunc(od) (pow(10,(c2 - od)))                                 //  OD units to RGB units
   
   Du = -0.01 * CMYKinputs[0];                                             //  convert inputs from cc to OD units
   Ru = -0.01 * CMYKinputs[1];                                             //  range is -0.99 to +0.99
   Gu = -0.01 * CMYKinputs[2];
   Bu = -0.01 * CMYKinputs[3];
   
   if (CMYKbias) {                                                         //  remove bias
      Du += 0.72;
      Gu += 0.32;
      Bu += 0.32;
   }
   
   con = -0.01 * CMYKinputs[4];
   conR = -0.01 * CMYKinputs[5];
   conG = -0.01 * CMYKinputs[6];
   conB = -0.01 * CMYKinputs[7];
   
   for (py = index; py < E3pxm->hh; py += Nwt)                             //  loop all image pixels
   for (px = 0; px < E3pxm->ww; px++)
   {
      if (sa_stat == 3) {                                                  //  select area active
         ii = py * E3pxm->ww + px;
         dist = sa_pixmap[ii];                                             //  distance from edge
         if (! dist) continue;                                             //  outside area
      }
      
      pix1 = PXMpix(E1pxm,px,py);                                          //  input pixel
      pix3 = PXMpix(E3pxm,px,py);                                          //  output pixel
      
      R1 = pix1[0] + 0.1;                                                  //  input RGB values, 0-255
      G1 = pix1[1] + 0.1;                                                  //  (avoid 0)
      B1 = pix1[2] + 0.1;
      
      Rd = ODfunc(R1);                                                     //  convert to OD units
      Gd = ODfunc(G1);
      Bd = ODfunc(B1);
      
      Rd += Du + Ru;                                                       //  add user OD increments
      Gd += Du + Gu;
      Bd += Du + Bu;
      
      if (con) {                                                           //  overall contrast
         Rd += con * c3 * (R1-128);
         Gd += con * c3 * (G1-128);
         Bd += con * c3 * (B1-128);
      }

      if (conR)                                                            //  single color contrast
         Rd += conR * c3 * (R1-128);
      if (conG) 
         Gd += conG * c3 * (G1-128);
      if (conB) 
         Bd += conB * c3 * (B1-128);
      
      R3 = RGBfunc(Rd);                                                    //  convert back to RGB units
      G3 = RGBfunc(Gd);
      B3 = RGBfunc(Bd);
      
      if (R3 < 0) R3 = 0;                                                  //  stop underflow
      if (G3 < 0) G3 = 0;
      if (B3 < 0) B3 = 0;

      if (R3 > 255.9 || G3 > 255.9 || B3 > 255.9) {                        //  stop overflow
         cmax = R3;
         if (G3 > cmax) cmax = G3;
         if (B3 > cmax) cmax = B3;
         F = 255.9 / cmax;
         R3 *= F;
         G3 *= F;
         B3 *= F;
      }

      if (sa_stat == 3 && dist < sa_blend) {                               //  select area is active,
         f1 = 1.0 * dist / sa_blend;                                       //    blend changes over sa_blend
         f2 = 1.0 - f1;
         R3 = f1 * R3 + f2 * R1;
         G3 = f1 * G3 + f2 * G1;
         B3 = f1 * B3 + f2 * B1;
      }
      
      pix3[0] = R3;                                                        //  output RGB values
      pix3[1] = G3;
      pix3[2] = B3;
   }

   exit_wthread();
   return 0;                                                               //  not executed, avoid gcc warning
}



